"""
This module provides configuration management for the Common-EGSE.

The configuration manager knows about the configuration of the system and the test setup. It's
main responsibility is to maintain the setup of the tests that are performed. It is the single
point access for all configuration information.

## The Configuration aka Setup

The Setup contains the identification of all the devices, mechanisms, controllers etc. that are
used for a particular test. For each of these items the Setup contains hardware and software
version, conversion and calibration information, location, components, specific settings,
defaults, in a word, all information that is needed to uniquely identify the components,
and to reproduce the test under the same circumstances. The details for the Setup are explained
in the `egse.setup` module.

## Prerequisites:

When the configuration manager (`cm_cs`) is started, no Setup is loaded. The configuration
manager will then be in a default Setup state without any devices configured. This is called the
_Zero_ Setup. The only keyword/attribute available from this Setup is `site_id`.

## Setup commands

The main purpose of the configuration manager is to maintain and manage Setups. These Setups
will reside in a GitHub repository for which the `cm_cs` has access to read and write Setups. The
`cm_cs` provides all configuration information on request with the following commands that
are available from the `ConfigurationManagerProxy`.

#### `list_setups()`

You can request a list of available Setups with the `list_setups` command. This function takes
keyword arguments which are the attributes of the Setup and compares the attribute with the
given value. An example should make this clear. A setup has a `site_id` and for the CSL site
also a `position`. You can access these value as follows:

    >>> from egse.setup import load_setup
    >>> setup = load_setup()
    >>> setup.site_id
    'CSL'
    >>> setup.position
    2

When you now want a list of all Setups specific for CSL that were applicable for position 2,
the following command will return that list.
```
with ConfigurationManagerProxy() as cm:
    print(cm.list_setups(site_id="CSL", position=2))
```
When you need to know which of these setups has the PUNA Hexapod with id=172543, add this
attribute as a keyword.
```
with ConfigurationManagerProxy() as cm:
    print(cm.list_setups(site_id="CSL", position=2, gse__hexapod__ID=172543))
```
When multiple attributes are specified, they are checked using a logical AND, not a logical OR,
meaning they have to meet _every_ attribute passed in and not just one of them.

You probably also noticed that instead of using the normal dot-notation to reach the hexapod id,
e.g. `gse.hexapod.ID`, we use double underscores to replace the dots. The reason for that is
that you can not have dots in keyword argument names. When you put a dot, you will get a
`SyntaxError`.

#### `load_setup(setup_id: int)`

Load a new Setup into the configuration manager. This command can only be called outside the
scope of an observation and will not have any effect when an observation is currently running.
Since the `cm_cs` knowns what the site_id is, the Setup for the current site is loaded
automativally.

#### `get_setup()`

Returns the Setup that is currently loaded on the configuration manager.


## Observation (aka Test) Commands

The configuration manager needs to know when an observation is started. It will keep track and
inform clients of the running observation.

#### `start_observation()`

This command starts a new observation. This will assign an new unique observation
identifier (`obsid`) for the observation and inform the Storage Manager that a new test has been
started with that `obsid`. A new test can not start before the previous test has been finished.
Also, a new Setup can not be loaded when an observation is running.

#### `end_observation()`

This command ends the current observation and notifies the Storage Manager that the test
has been ended.

#### `get_obsid()`

Returns the observation identifier of the currently running observation.


## Convenience Functions

The package also defines a number of convenience functions that simplify the communication
with the configuration manager `cm_cs`.

#### `is_configuration_manager_active()`

A function that checks if the `cm_cs` is running and responding to commands. This function makes
a connection with the `cm_cs` and sends it a _Ping_ command. This is the recommended way to check
the availability of the configuration manager.

"""

from __future__ import annotations

import operator
import random
import textwrap
import threading
import time
from pathlib import Path
from typing import NamedTuple
from typing import Optional
from typing import Union

import git
import rich
from git import GitCommandError

from egse.command import ClientServerCommand
from egse.command import stringify_function_call
from egse.config import find_file
from egse.config import find_files
from egse.connect import get_endpoint
from egse.control import ControlServer
from egse.control import is_control_server_active
from egse.decorators import dynamic_interface
from egse.decorators import static_vars
from egse.env import get_conf_data_location
from egse.env import get_conf_repo_location
from egse.env import get_project_name
from egse.env import get_site_id
from egse.exceptions import InternalError
from egse.log import logger
from egse.notifyhub.event import NotificationEvent
from egse.notifyhub.services import EventPublisher
from egse.obsid import ObservationIdentifier
from egse.plugin import entry_points
from egse.protocol import CommandProtocol
from egse.proxy import Proxy
from egse.registry.client import RegistryClient
from egse.response import Failure
from egse.response import Response
from egse.response import Success
from egse.settings import Settings
from egse.settings import SettingsError
from egse.setup import Setup
from egse.setup import disentangle_filename
from egse.setup import load_last_setup_id
from egse.setup import save_last_setup_id
from egse.system import duration
from egse.system import filter_by_attr
from egse.system import format_datetime
from egse.system import humanize_seconds
from egse.version import get_version_installed
from egse.zmq_ser import bind_address
from egse.zmq_ser import connect_address

HERE = Path(__file__).parent

settings = Settings.load("Configuration Manager Control Server")

SITE_ID = get_site_id()
COMMAND_SETTINGS = Settings.load(location=HERE, filename="confman.yaml")

PROCESS_NAME = settings.get("PROCESS_NAME", "cm_cs")
PROTOCOL = settings.get("PROTOCOL", "tcp")
HOSTNAME = settings.get("HOSTNAME", "localhost")
COMMANDING_PORT = settings.get("COMMANDING_PORT", 0)
SERVICE_PORT = settings.get("SERVICE_PORT", 0)
MONITORING_PORT = settings.get("MONITORING_PORT", 0)
STORAGE_MNEMONIC = settings.get("STORAGE_MNEMONIC", "CM")
SERVICE_TYPE = settings.get("SERVICE_TYPE", "cm_cs")

# CM_SETUP_ID = Gauge("CM_SETUP_ID", 'Setup ID')
# CM_TEST_ID = Gauge("CM_TEST_ID", 'Test ID')

PROXY_TIMEOUT = 10.0  # don't wait longer than 10s by default


def _push_setup_to_repo(filename: str, commit_msg: str) -> Failure | Success:
    """
    Push the Setup file to the `cgse-conf` repository on GitHub.

    Args:
        filename: the basename of the new Setup file

    Returns:
        None.
    """

    repo_workdir = get_conf_repo_location()

    if repo_workdir is None:
        msg = textwrap.dedent(
            f"""\
            Couldn't determine the repository location for configuration data. 

            Check if the environment variable '{get_project_name()}_CONF_REPO_LOCATION' is set and valid 
            before starting the configuration manager.
            """
        )
        logger.error(msg)
        return Failure(msg)

    repo = git.Repo(repo_workdir)

    if repo.is_dirty():
        logger.warning(f"The cgse-conf repository is dirty. Check the git status at '{repo_workdir}'.")

    untracked = repo.untracked_files

    if len(untracked) != 1:
        msg = textwrap.dedent(
            f"""\
            The number of untracked files ({len(untracked)}) in the cgse-conf repository doesn't match 
            the expected. Check the git status at '{repo_workdir}' on the egse-server. 
            Only '{filename}' should be untracked.
            
            Untracked files: {untracked}
            """
        )
        logger.error(msg)
        return Failure(msg)

    # match the filename to extract the full path to the file

    untracked = [x for x in untracked if filename in x]
    if (n := len(untracked)) != 1:
        msg = f"There should be one match for the filename, found {n}{'' if n == 0 else untracked}."
        logger.error(msg)
        return Failure(msg)

    untracked = untracked[0]

    # The response is a list of tuples containing the path of the file added to the stages/index.

    try:
        response = repo.index.add(untracked)
        # assert response[0].path == untracked
    except FileNotFoundError:
        # if for some reason the untracked file can not be found, should not happen..
        logger.warning(f"Untracked file {untracked} not found. Check the git status at {repo_workdir}.")

    # The response is a Commit object containing e.g. the commit message, the hash, the author, ...

    response = repo.index.commit(message=commit_msg)

    # The response is a list of FetchInfo instances
    # We need this `pull` command before we will push the changes because otherwise the push will
    # be rejected, see https://github.com/IvS-KULeuven/plato-common-egse/issues/2027. The problem
    # should not abort the submit command, but needs to be logged.

    try:
        response = repo.remote("upload").pull("main")
    except Exception as exc:
        logger.error(exc, exc_info=True)

    # The response is a PushInfo object

    try:
        response = repo.remote("upload").push("main")
    except ValueError as exc:
        logger.error(exc, exc_info=True)
        return Failure(f"Push of setup [{filename}] failed", exc)
    except GitCommandError as exc:
        logger.error(exc, exc_info=True)
        return Failure(f"Push of setup [{filename}] failed", exc)

    return Success(f"Successfully pushed the setup to the repo {repo}")


# We have seen that especially when listing the setups, we have a performance problem.
# Therefore, we implement a cache for the Setup info that we use in different functions.
#
# The key is the Setup ID
# The value is the named tuple SetupInfo

_cached_setup_info = {}


class SetupInfo(NamedTuple):
    path: Path
    site_id: str
    sut_id: str
    description: str


def _populate_cached_setup_info():
    """
    Populates the internal cache of Setup information.

    Raises:
        InternalError when a Setup is loaded that doesn't have an ID associated.

    """
    global _cached_setup_info

    logger.info("Populating cache with Setup Info.")

    location = get_conf_data_location()
    if location:
        data_conf_location = Path(location).expanduser()
    else:
        raise ValueError(
            "Couldn't determine location of the configuration data with 'get_conf_data_location()'. "
            "Check if the environment is properly defined."
        )

    setup_info = {}

    for fn in find_files(pattern="SETUP*", root=data_conf_location):
        setup = Setup.from_yaml_file(fn)
        if id := setup.get_id():
            id = int(id)
            site_id = _get_site_id_for_setup(setup)
            sut_id = _get_sut_id_for_setup(setup)
            description = _get_description_for_setup(setup)
            setup_info[id] = SetupInfo(fn, site_id, sut_id, description)
        else:
            raise InternalError(f"Setup loaded without an ID, {fn=}")

        time.sleep(0.1)

    _cached_setup_info = dict(sorted(setup_info.items()))

    logger.info("SetupInfo cache populated.")


def _add_setup_info_to_cache(setup: Setup):
    global _cached_setup_info

    if (_id := setup.get_id()) is None:
        raise InternalError(f"Setup loaded without an ID, {setup=!s}")

    if (_fn := setup.get_filename()) is None:
        raise InternalError(f"Setup with id={_id} has no filename associated.")

    _id = int(_id)
    _fn = Path(_fn)

    site_id = _get_site_id_for_setup(setup)
    sut_id = _get_sut_id_for_setup(setup)
    description = _get_description_for_setup(setup)

    _cached_setup_info[_id] = SetupInfo(_fn, site_id, sut_id, description)


def _print_cached_setup_info():
    global _cached_setup_info

    rich.print(_cached_setup_info)


def _get_cached_setup_info(setup_id: int) -> Optional[SetupInfo]:
    """Returns a setup info named tuple for the given setup id or None when no
    SetupInfo for the given setup_id is available.."""
    global _cached_setup_info

    return _cached_setup_info.get(setup_id)


def _reload_cached_setup_info():
    try:
        Setup.from_yaml_file.cache_clear()
    except AttributeError:
        logger.warning("Setup.from_yaml_file() method is not decorated with an lru_cache.")

    _populate_cached_setup_info()


def is_configuration_manager_active(timeout: float = 0.5):
    """
    Checks whether the Configuration Manager is running.

    Args:
        timeout (float): Timeout when waiting for a reply [seconds, default=0.01]

    Returns:
        True if the Configuration Manager is running and replied with the expected answer.
    """

    if COMMANDING_PORT == 0:
        with RegistryClient() as client:
            endpoint = client.get_endpoint(settings.SERVICE_TYPE)
            if endpoint is None:
                logger.debug(f"No endpoint for {settings.SERVICE_TYPE}")
                return False
    else:
        endpoint = connect_address(PROTOCOL, HOSTNAME, COMMANDING_PORT)

    return is_control_server_active(endpoint, timeout)


def _construct_filename(site_id: str, setup_id: int, creation_time: str = None):
    """Construct a filename for a Setup.

    FIXME: describe the restrictions on file naming and how they are parsed etc.

    Args:
        site_id (str): the site identifier
        setup_id (int): the setup identifier
        creation_time (str): the date-time shall be formatted as `YYMMDD_HHMMSS`
    """

    if creation_time:
        filename = f"SETUP_{site_id}_{setup_id:05d}_{creation_time}.yaml"
    else:
        filename = f"SETUP_{site_id}_{setup_id:05d}_{format_datetime(fmt='%y%m%d_%H%M%S')}.yaml"

    return filename


def _get_description_for_setup(setup: Setup, setup_id: int = None) -> str:
    setup_id = setup_id or int(setup.get_id())
    try:
        description = setup.history.get(setup_id)
    except AttributeError:
        description = None
    return description or f"no description found for Setup {setup_id}"


def _get_sut_id_for_setup(setup: Setup) -> str | None:
    try:
        if "id" in setup.sut:
            sut_id = setup.sut.id.lower()
        elif "ID" in setup.sut:
            sut_id = setup.sut.ID.lower()
        else:
            sut_id = None
    except AttributeError:
        sut_id = None

    return sut_id


def _get_site_id_for_setup(setup: Setup) -> str:
    try:
        site_id = setup.site_id if "site_id" in setup else None
    except AttributeError:
        site_id = None

    return site_id or "no site_id"


@static_vars(test_id=0)
def create_obsid(last_obsid: str, site_id: str, setup_id: int):
    # This is method is currently just a prove of concept, the real thing should
    # read the LID, SID from the current Setup and then generate or keep track
    # of a TEST ID.

    # How do we guarantee a unique OBSID? OBSIDs need to be made persistent at least for the Site.
    # That way we can, when a new ObservationIdentifier is generated, check if it's indeed unique.

    if last_obsid:
        create_obsid.test_id = int(last_obsid.split(maxsplit=1)[0])

    create_obsid.test_id += 1

    # We need access to the setup, because we need the LabID, the SetupID
    # How do we define the configuration ID and Test ID?

    return ObservationIdentifier(site_id, setup_id, create_obsid.test_id)


class ConfigurationManagerInterface:
    """
    This interface is for sending commands to the configuration manager to e.g. start and stop
    an observation/test, or get information about the Setups.

    The interface should be implemented by the `ConfigurationManagerController` and the
    `ConfigurationManagerProxy` (and possibly a `ConfigurationManagerSimulator` should we
    need that e.g. for testing purposes).
    """

    @dynamic_interface
    def start_observation(self, function_info: dict) -> Response:
        """Starts a new observation or test. The following actions will be taken:

        * create an observation identifier, aka `obsid`
        * notify the Storage Manager Control Server that a new observation is started
        * return the generated `obsid`

        Args:
            function_info: dictionary with information about the function called
        Returns:
            `Success` (with `obsid` as `return_code`) or `Failure` when already in an observation
            or Storage returned Failure.
        """
        raise NotImplementedError

    @dynamic_interface
    def end_observation(self) -> Response:
        """Ends the current observation and notifies the Storage Manager Control Server.

        Returns:
            `Success` when the observation could be closed properly and the Storage CS was notified
            or `Failure` otherwise.
        """
        raise NotImplementedError

    @dynamic_interface
    def get_obsid(self) -> Success:
        """Returns the current observation identifier. When no observation is running, `None` is
        returned as the `return_code` in `Success`.

        Returns:
            Always returns `Success` with current observation identifier, i.e. `obsid`.
        """
        raise NotImplementedError

    @dynamic_interface
    def register_to_storage(self):
        """Register the configuration manager to the Storage manager.

        Registration is done for `obsid` and `CM` which handle different purposes:

        - `obsid` for accessing the OBSID table
        - 'CM' for storing housekeeping information from the configuration manager

        """
        raise NotImplementedError

    @dynamic_interface
    def load_setup(self, setup_id: int = None) -> Union[Setup, Failure]:
        raise NotImplementedError

    @dynamic_interface
    def get_setup(self, setup_id: int = None):
        raise NotImplementedError

    @dynamic_interface
    def reload_setups(self):
        raise NotImplementedError

    @dynamic_interface
    def list_setups(self, **attr):
        raise NotImplementedError

    @dynamic_interface
    def submit_setup(self, setup: Setup, description: str, replace: bool = True) -> Setup | None:
        raise NotImplementedError

    @dynamic_interface
    def get_setup_for_obsid(self, obsid):
        raise NotImplementedError


class ConfigurationManagerController(ConfigurationManagerInterface):
    """Handles the commands that are sent to the configuration manager.

    .. note::
        The docstrings for each of the commands are in the `ConfigurationManagerInterface`.
    """

    def __init__(self, control_server: ControlServer | None = None):
        self._obsid: ObservationIdentifier | None = None
        self._obsid_start_dt: str | None = None
        self._setup: Setup | None = None
        self._setup_id: int | None = None
        self._sut_id: str | None = None
        self._control_server: ControlServer | None = control_server

        self.register_to_storage()

        # Find the location for the configuration data

        location = get_conf_data_location()
        if location:
            self._data_conf_location = Path(location).expanduser()
        else:
            raise ValueError("The location for the configuration data is not defined. Please check your environment.")

        # Populate the cache with information from the available Setups. This will also load each
        # Setup and cache them with the lru_cache decorator. Since this takes about 5s for 100
        # Setups, we run this function in a daemon thread in order not to block the cm_cs from
        # reacting to command requests.

        cache_thread = threading.Thread(target=_populate_cached_setup_info)
        cache_thread.daemon = True
        cache_thread.start()

        # Load the last used Setup

        setup_id = load_last_setup_id()
        self.load_setup(setup_id)

    def quit(self):
        if self._storage:
            self._storage.disconnect_cs()

    @property
    def data_location(self) -> Path:
        """Return the location of the configuration data, i.e. the Setup YAML files."""
        return self._data_conf_location

    def start_observation(self, function_info: dict) -> Response:
        if self._obsid is not None:
            return Failure(
                "An new observation can not be started before the previous observation is "
                "finished. You will need to first send an end_observation request to the "
                "configuration manager."
            )

        last_obsid = None

        if self._storage:
            last_obsid = self._storage.read({"origin": "obsid", "select": "last_line"})
            last_obsid = last_obsid.return_code if isinstance(last_obsid, Success) else None

        self._obsid = create_obsid(last_obsid, SITE_ID, self._setup_id)
        self._obsid_start_dt = format_datetime()

        if self._storage:
            response = self._storage.start_observation(self._obsid, self._sut_id)
        else:
            return Failure("Couldn't send start observation to Storage Manager, no Storage Manager available.")

        if not response.successful:
            self._obsid = None
            return Failure(
                "Sending a start_observation to the Storage Manager Control Server failed",
                response,
            )

        description = function_info.pop("description", "")
        cmd = stringify_function_call(function_info).replace("\n", " ")

        if description:
            cmd += f" [{description}]"

        response = self._storage.save(
            {
                "origin": "obsid",
                "data": f"{self._obsid.test_id:05d} "
                f"{self._obsid.lab_id} "
                f"{self._obsid.setup_id:05d} "
                f"{self._obsid_start_dt} "
                f"{cmd}",
            }
        )

        if isinstance(response, Failure):
            logger.warning(f"There was a Failure when saving to the obsid-table: {response}")
        else:
            logger.info(f"Successfully created an observation with obsid={self._obsid}.")

        return Success("Returning the OBSID", self._obsid)

    def end_observation(self) -> Response:
        if not self._obsid:
            return Failure("Received end_observation command while not currently in an observation context.")

        if self._storage:
            response = self._storage.end_observation(self._obsid)
        else:
            return Failure("Couldn't send end observation to Storage Manager, no Storage Manager available.")

        if not response.successful:
            return Failure(
                "Sending an end_observation to the Storage Manager Control Server failed.",
                response,
            )

        obsid_end_dt = format_datetime()
        obs_duration = humanize_seconds(
            duration(self._obsid_start_dt, obsid_end_dt).total_seconds(), include_micro_seconds=False
        )
        logger.info(f"Successfully ended observation with obsid={self._obsid}, duration={obs_duration}.")

        self._obsid = None
        self._obsid_start_dt = None

        return Success("Successfully ended the observation.")

    def get_obsid(self) -> Success:
        if self._obsid:
            msg = "Returning the current OBSID."
        else:
            msg = "No observation running. Use start_observation() to start an observation."
        return Success(msg, self._obsid)

    def register_to_storage(self):
        # Import these modules here as to optimize the import of classes and functions in other parts of the CGSE.
        # The ConfigurationManagerController is only used by the CM CS and these Storage imports are only used in
        # this class and take too much loading time...

        from egse.storage import StorageProxy
        from egse.storage import is_storage_manager_active
        from egse.storage.persistence import TYPES

        if is_storage_manager_active():
            self._storage = StorageProxy()
            response = self._storage.register(
                {
                    "origin": "obsid",
                    "persistence_class": TYPES["TXT"],
                    "prep": {"mode": "a", "ending": "\n"},
                    "persistence_count": True,
                    "filename": "obsid-table.txt",
                }
            )
            logger.info(response)
        else:
            self._storage = None
            logger.error("No Storage Manager available !!!!")

    def load_setup(self, setup_id: int = None) -> Union[Setup, Failure]:
        """Load the Setup with the given setup_id.

        Args:
            setup_id (int): the identifier for the requested Setup.
        Returns:
            The requested Setup.
        """
        # The current implementation is file based. The files have a strict naming convention and
        # are located in the `data/conf` directory.
        #
        # 1. Find the Setup for the given setup_id and the Site (is this read from the Settings
        #    file, or set by some GUI or process?
        # 2. Load that Setup from its file at the default location
        # 3. Return an acknowledgement that the Setup is loaded on the CM_CS or not

        if setup_id is None:
            return Failure(
                "No Setup ID was given, cannot load a Setup into the configuration manager. "
                "If you wanted to get the current Setup from the configuration manager, use the "
                "get_setup() method instead."
            )

        if self._obsid:
            return Failure(
                f"A new Setup can not be loaded when an observation is running. "
                f"The current obsid is {self._obsid}. Use `end_observation()` before "
                f"loading a new Setup."
            )

        setup_files = list(find_files(pattern=f"SETUP_{SITE_ID}_{setup_id:05d}_*.yaml", root=self._data_conf_location))

        if len(setup_files) != 1:
            logger.error(
                msg := f"Expected to find just one Setup YAML file, found {len(setup_files)}. "
                f"[{SITE_ID = }, {setup_id = }, data_conf_location={self._data_conf_location}]"
            )
            return Failure("Loading Setup", InternalError(msg))

        setup_file = setup_files[0]

        try:
            self._setup = Setup.from_yaml_file(setup_file)
            self._setup_id = setup_id
            self._sut_id = _get_sut_id_for_setup(self._setup)
            logger.info(f"New Setup loaded from {setup_file}")
            save_last_setup_id(self._setup_id)

            with EventPublisher() as pub:
                pub.publish(
                    NotificationEvent(event_type="new_setup", source_service="cm_cs", data={"setup_id": self._setup_id})
                )

            return self._setup
        except SettingsError as exc:
            return Failure(f"The Setup file can not be loaded from {setup_file}.", exc)

    def get_setup(self, setup_id: int = None) -> Union[Setup, Failure]:
        """
        Returns the Setup for the given setup_id. If no setup_id argument was provided,
        the current Setup from the configuration manager will be returned.

        This is a -read-only function.
        Under no circumstance will the Setup of the configuration manager be changed.

        Args:
            setup_id (int): the identifier for the requested Setup.
        Returns:
            The requested Setup.
        """

        if setup_id:
            # If a Setup ID is given, just load and return the Setup for that ID
            # The Setup is NOT added to the Configuration Manager as the current Setup.

            setup_files = list(
                find_files(pattern=f"SETUP_{SITE_ID}_{setup_id:05d}_*.yaml", root=self._data_conf_location)
            )

            if len(setup_files) != 1:
                logger.error(msg := f"Expected to find just one Setup YAML file, found {len(setup_files)}.")
                return Failure("Expected only one Setup.", InternalError(msg))

            setup_file = setup_files[0]

            try:
                return Setup.from_yaml_file(setup_file)
            except SettingsError as exc:
                return Failure(f"The Setup file can not be loaded from {setup_file}.", exc)

        else:
            # No Setup ID was given, so we return the current Setup loaded in the Configuration Manager

            if self._setup:
                return self._setup
            else:
                return Failure("No Setup was loaded on the Configuration Manager.")

    def get_setup_id(self) -> int:
        """Returns the Setup identifier for the currently loaded Setup.

        Returns:
            The `setup_id` of the Setup loaded by the `cm_cs`, or None.
        """

        return self._setup_id

    def get_site_id(self) -> str:
        """Returns the Site identifier that is used by the configuration manager.

        Returns:
            The Site identifier as a string.
        """

        return SITE_ID

    def reload_setups(self):
        """
        Clears the cache and Reloads the available Setups.

        This function does not affect the currently loaded Setup.
        """
        _reload_cached_setup_info()

    def list_setups(self, **attr):
        """
        Returns a sorted list of all the available Setups for the current site. The list contains
        tuples with the following content: setup_id, site_id, description, cam_id.

        Args:
            **attr: see egse.system.filter_by_attr()

        Returns:
            A list with information on the available Setups.
        """
        # The current implementation is file based. The files have a strict naming convention and
        # are located in the `data/conf` directory.
        #
        # 1. Get a list of the Setup files from the default location, i.e. data/conf
        # 2. Prepare a list of tuples with that information ordered by Setup ID
        # 3. Return that list

        setup_list = []

        setups = [Setup.from_yaml_file(info.path) for info in _cached_setup_info.values()]

        setups = filter_by_attr(setups, **attr)

        for setup in setups:
            # FIXME: are we sure setup.get_filename() returns a Path?
            setup_site, setup_id = disentangle_filename(str(setup.get_filename()))
            description = _get_description_for_setup(setup, int(setup_id))
            sut_id = _get_sut_id_for_setup(setup)
            setup_list.append((setup_id, setup_site, description, sut_id))

        # Sort by site, then by id

        return sorted(setup_list, key=operator.itemgetter(1, 0), reverse=False)

    def get_setup_for_obsid(self, obsid):
        obsid = f"{obsid:05d}" if isinstance(obsid, int) else obsid
        rc = self._storage.read({"origin": "obsid", "select": ("startswith", obsid)})
        if rc.successful:
            # FIXME: this should be a function that can also be used in load_setup(),
            #  because they do basically the same thing
            try:
                setup_id = int(rc.return_code[-1].split(maxsplit=3)[2])
                setup_file = find_file(name=f"SETUP_{SITE_ID}_{setup_id:05d}_*.yaml", root=self._data_conf_location)
                setup = Setup.from_yaml_file(setup_file)
            except (IndexError, SettingsError):
                setup = None

        return setup

    def submit_setup(self, setup: Setup, description: str, replace: bool = True):
        # 1. Determine the Site for this Setup, or should this be the Site that is known by the
        #    CM_CS?
        # 2. Find the correct (next) number for the Setup for the given Site
        # 3. Do I want to make some comparison?
        #    Do we need to keep a record from which this Setup is derived?

        # FIXME: define and handle exceptional conditions, like IOError

        if self._obsid is not None:
            return Failure(
                "An new Setup can not be submitted when an observation is running. You will need "
                "to first send an end_observation request to the configuration manager."
            )

        site = setup.site_id

        setup_id = self.get_next_setup_id_for_site(site)

        filename = _construct_filename(SITE_ID, setup_id)

        if not hasattr(setup, "history"):
            setup.history = {}

        setup.history.update({f"{setup_id}": description})
        setup.set_private_attribute("_setup_id", setup_id)
        setup.to_yaml_file(self._data_conf_location / filename)

        # No repository is defined. This should not break, but a warning is in place.
        # The warnings are issued by the get_conf_repo_location() function.

        if get_conf_repo_location():
            try:
                rc = _push_setup_to_repo(filename, description)
                if isinstance(rc, Failure):
                    return rc
                _add_setup_info_to_cache(setup)
            except (Exception,) as exc:
                msg = "Submit_setup could not complete it's task to send the new Setup to the repo."
                logger.error(msg, exc_info=True)
                return Failure("Submit_setup could not complete it's task to send the new Setup to the repo.", exc)
            else:
                logger.info(f"Successfully pushed Setup {setup_id} to the repository.")

        if replace:
            self._setup = setup
            self._setup_id = setup_id
            logger.info(f"New Setup was submitted and loaded: {setup_id=}")
            self._sut_id = _get_sut_id_for_setup(setup)
            save_last_setup_id(setup_id)

            with EventPublisher() as pub:
                pub.publish(
                    NotificationEvent(event_type="new_setup", source_service="cm_cs", data={"setup_id": self._setup_id})
                )

        return setup

    def get_next_setup_id_for_site(self, site: str) -> int:
        """
        Return the next available Setup ID for the given Site.

        Args:
            site (str): site identification, e.g. CSL, SRON, VACUUM_LAB...
        """
        site = site or SITE_ID
        files = sorted(find_files(pattern=f"SETUP_{site}_*.yaml", root=self._data_conf_location))
        last_file = files[-1]
        _, setup_id = disentangle_filename(last_file.name)

        return int(setup_id) + 1


class ConfigurationManagerCommand(ClientServerCommand):
    pass


class ConfigurationManagerProxy(Proxy, ConfigurationManagerInterface):
    """
    The Configuration Manager Proxy class is used to connect to the Configuration Manager
    Control Server and send commands and requests for the configuration manager.

    When the port number passed is 0 (zero), the endpoint is requested from the
    service registry.

    Args:
        protocol: the transport protocol [default is taken from settings file]
        hostname: location of the control server (IP address) [default is taken
            from settings file]
        port: TCP port on which the control server is listening for commands
            [default is taken from settings file]
        timeout: number of fractional seconds before a timeout is triggered
    """

    def __init__(
        self, protocol: str = PROTOCOL, hostname: str = HOSTNAME, port: int = COMMANDING_PORT, timeout=PROXY_TIMEOUT
    ):
        endpoint = get_endpoint(settings.SERVICE_TYPE, protocol, hostname, port)

        super().__init__(endpoint, timeout=timeout)


class ConfigurationManagerProtocol(CommandProtocol):
    def __init__(self, control_server: ControlServer):
        super().__init__(control_server)

        self.controller = ConfigurationManagerController(control_server)

        self.load_commands(
            COMMAND_SETTINGS.Commands,
            ConfigurationManagerCommand,
            ConfigurationManagerController,
        )

        self.build_device_method_lookup_table(self.controller)

        self.version_dict = {}
        for ep in sorted(entry_points("cgse.version"), key=lambda x: x.name):
            if installed_version := get_version_installed(ep.name):
                self.version_dict[f"CM_{ep.name.upper().replace('-', '_')}"] = installed_version

    def get_bind_address(self):
        return bind_address(
            self.control_server.get_communication_protocol(),
            self.control_server.get_commanding_port(),
        )

    def get_status(self) -> dict:
        status = super().get_status()

        status.update({"obsid": self.controller.get_obsid().return_code})
        status.update({"setup": self.controller.get_setup()})

        return status

    def get_housekeeping(self) -> dict:
        obsid = self.controller.get_obsid().return_code
        test_id = obsid.test_id if obsid else float("nan")
        setup_id = self.controller.get_setup_id()
        site_id = self.controller.get_site_id()

        hk = {
            "timestamp": format_datetime(),
            "random": random.randint(0, 100),
            "CM_SITE_ID": site_id,
            "CM_SETUP_ID": setup_id,
            "CM_TEST_ID": test_id,
            "CM_OBSID": obsid,
        }

        hk.update(self.version_dict)

        return hk

    def quit(self):
        self.controller.quit()


# The following functions are defined here to allow them to be used in the list_setups() method
# and be pickled and passed over ZeroMQ.


def is_in(a, b):
    """Returns result of `a in b`."""
    return a in b


def is_not_in(a, b):
    """Returns result of `a not in b`."""
    return a not in b


def get_status():
    try:
        with ConfigurationManagerProxy() as cm:
            obsid = cm.get_obsid()
            obsid = obsid.return_code
            setup = cm.get_setup()
            try:
                site_id = f"Site ID: {setup.site_id}"
            except AttributeError:
                site_id = "Site ID: [red]UNKNOWN[/]"

            if obsid:
                obsid = f"Running observation: {obsid}"
            else:
                obsid = "No observation running"

            try:
                if setup.has_private_attribute("_setup_id"):
                    setup_id = setup.get_private_attribute("_setup_id")
                    setup_id = f"Setup loaded: {setup_id}"
                else:
                    setup_id = "[red]No Setup loaded[/]"
            except Exception as exc:
                setup_id = f"An Exception was caught: {exc}"

            return textwrap.dedent(
                f"""\
                Configuration manager:
                    Status: [green]active[/]
                    {site_id}
                    {obsid}
                    {setup_id}
                    Hostname: {cm.get_ip_address()}
                    Monitoring port: {cm.get_monitoring_port()}
                    Commanding port: {cm.get_commanding_port()}
                    Service port: {cm.get_service_port()}
                """
            )
    except ConnectionError as exc:
        return f"Configuration Manager Status: [red]not active[/] ({exc})"



from __future__ import annotations

import asyncio
import logging
import random
import sys
import threading
from typing import Any
from typing import Callable

try:
    from typing import override
except ImportError:
    try:
        from typing_extensions import override
    except ImportError:

        def override(method, /):
            return method


import typer

from egse.async_control import AcquisitionAsyncControlServer
from egse.async_control import TypedAsyncControlClient
from egse.log import logger
from egse.logger import remote_logging
from egse.system import TyperAsyncCommand
from egse.system import do_every
from egse.zmq_ser import zmq_json_response
from egse.zmq_ser import zmq_string_response


class DummyDAQSimulator:
    """Minimal DAQ-like simulator that invokes a callback from a dedicated thread.

    This simulator produces records with a simple sequence number and random value at a configurable interval.
    It can be instantiated and started/stopped multiple times to test acquisition lifecycle management
    in the control server. The acquisition callback is invoked in a try/except block to prevent thread crashes
    from exceptions in the callback, which can help with debugging issues in the control server without
    losing the simulator thread.

    An example use can be found in the `start-acquisition` command handler of `DummyAsyncControlServer`,
    which creates and starts the simulator, and in the `stop_acquisition` method, which stops it.
    """

    def __init__(self, callback: Callable[..., None], interval_s: float):
        """Create a thread-backed producer.

        Args:
            callback: Acquisition callback provided by the control server.
            interval_s: Time between produced records.
        """
        self._callback = callback
        self._interval_s = interval_s
        self._running = threading.Event()
        self._stop = threading.Event()
        self._thread: threading.Thread | None = None
        self._sequence = 0
        self.logger = logging.getLogger("egse.dummy_daq_simulator")

    @property
    def running(self) -> bool:
        """Return True when the simulator thread is alive and acquisition is enabled."""
        return self._running.is_set() and self._thread is not None and self._thread.is_alive()

    def start(self):
        """Start the simulator thread if it is not already running."""
        if self.running:
            return
        self._running.set()
        self._thread = threading.Thread(target=self._run, daemon=True, name="dummy-daq-simulator")
        self._thread.start()

    def stop(self):
        """Stop the simulator thread and wait briefly for it to exit."""
        self._running.clear()
        self._stop.set()
        thread = self._thread
        if thread is not None:
            thread.join(timeout=1.0)
        self._thread = None

    def _callback_wrapper(self):
        """Invoke the acquisition callback and catch any exceptions to prevent thread crashes.

        This simulates a device driver invoking a callback for each produced record of device information/data.
        The callback is expected to be provided by the control server and can perform any processing or forwarding
        of the produced record. Wrapping the callback in a try/except block allows us to log any exceptions that
        occur without crashing the simulator thread, which can help with debugging issues in the control server's
        acquisition handling.

        The callback is expected to accept at least one argument which can be anything, but in this case we provide a
        dictionary with a sequence number and random value to simulate produced records of device data.
        The callback can also accept additional keyword arguments for metadata, source, etc. if desired.
        """
        try:
            self._sequence += 1
            self._callback(
                {
                    "sequence": self._sequence,
                    "value": random.random(),
                },
                source="dummy-acquisition",
                metadata={"interval": self._interval_s, "origin": "dummy-daq-simulator"},
            )
        except Exception as exc:
            # Log exceptions from the callback to help debug issues without crashing the thread
            self.logger.error(f"Exception in DummyDAQSimulator callback: {type(exc).__name__}: {exc}")

    def _run(self):
        do_every(self._interval_s, self._callback_wrapper, stop_event=self._stop)


class DummyAsyncControlServer(AcquisitionAsyncControlServer):
    """Example async control server showing what belongs in a subclass.

    The superclass handles sockets, request/response framing, registration, lifecycle,
    task management, and common service commands. This subclass only defines
    device-specific commands and extra service metadata.
    """

    service_type = "dummy-async-control-server"

    def __init__(self):
        self._last_value: str | None = None
        self._echo_count = 0
        self._acquisition_logged_count = 0
        self._acquisition_interval_s = 0.05
        self._daq_simulator: DummyDAQSimulator | None = None
        super().__init__()

    @override
    def register_custom_handlers(self):
        """Register dummy command handlers, including acquisition lifecycle commands."""
        self.add_device_command_handler("echo", self._do_echo)
        self.add_device_command_handler("set-value", self._do_set_value)
        self.add_device_command_handler("start-acquisition", self._do_start_acquisition)
        self.add_device_command_handler("stop-acquisition", self._do_stop_acquisition)
        self.add_service_command_handler("health", self._handle_health)

    @override
    def get_service_info(self) -> dict[str, Any]:
        """Return service metadata, including acquisition state and counters."""
        info = super().get_service_info()
        info.update(
            {
                "supports": [
                    "echo",
                    "set-value",
                    "start-acquisition",
                    "stop-acquisition",
                    "health",
                    "ping",
                    "info",
                    "terminate",
                ],
                "echo count": self._echo_count,
                "last value": self._last_value,
                "acquisition running": self._is_acquisition_running(),
                "acquisition logged": self._acquisition_logged_count,
                "acquisition interval": self._acquisition_interval_s,
            }
        )
        return info

    async def _do_echo(self, cmd: dict[str, Any]) -> list:
        """Echo the provided message and increment a call counter."""
        payload = str(cmd.get("message", ""))
        self._echo_count += 1
        return zmq_string_response(payload)

    async def _do_set_value(self, cmd: dict[str, Any]) -> list:
        """Store a value in memory so tests can verify device state transitions."""
        self._last_value = str(cmd.get("value", ""))
        return zmq_json_response({"success": True, "message": {"stored": self._last_value}})

    def _is_acquisition_running(self) -> bool:
        """Return True when the DAQ simulator is actively producing samples."""
        return self._daq_simulator is not None and self._daq_simulator.running

    def custom_callback(self, data, *args, **kwargs):
        # Do some custom processing of the device driver record here, this might be to extract data from the record, but
        # the 'record' might also be a handle to a device driver object that can be queried for more information, or it
        # could be any other data structure depending on how the acquisition callback is invoked by the device
        # driver/simulator. In this example, we just log the record sequence and value for demonstration purposes,
        # but in a real control server this is where you would implement the logic to handle incoming acquisition
        # records, such as parsing the data, applying calibrations, checking for alerts, forwarding to other services,
        # etc.

        # The DummyDAQSimulator provides a dictionary as the record, but this could be any data structure.
        record = data
        logger.debug(f"Custom processing of record {record.get('sequence')} with value {record.get('value')}")

        # Then pass it to the default callback for further processing and finally enqueuing in the acquisition queue.
        self.on_acquisition_data(record, source="custom", metadata={"processed": True})

    async def _do_start_acquisition(self, cmd: dict[str, Any]) -> list:
        """Start the DAQ simulator and stream samples through the acquisition callback."""
        interval_s = float(cmd.get("interval", self._acquisition_interval_s))
        if interval_s <= 0:
            return zmq_json_response(
                {
                    "success": False,
                    "message": {"error": "interval must be > 0"},
                }
            )

        self._acquisition_interval_s = interval_s

        # Don't start a new simulator if one is already running, but return success with the current state
        if self._is_acquisition_running():
            return zmq_json_response(
                {
                    "success": True,
                    "message": {
                        "running": True,
                        "interval": self._acquisition_interval_s,
                        "already_running": True,
                    },
                }
            )

        # This will return the default acquisition callback function that is defined in the superclass.
        # The default callback expects one positional argument for the acquisition record, which can be any data
        # structure but in this case we provide a dictionary with a sequence number and random value to simulate
        # produced records of device data. The default callback also accept optional keyword arguments,
        # in this case, 'source' and 'metadata' are provided.

        callback = self.get_acquisition_callback()

        # Alternatively, you could also define a custom callback function here in the control server subclass that
        # performs additional processing or forwarding of the produced records, and pass that to the device
        # acquisition method instead of the default callback from the superclass.
        # For example:

        # callback = self.custom_callback

        # The DummyDAQSimulator runs in a separate thread and invokes the callback for each produced record.
        self._daq_simulator = DummyDAQSimulator(callback=callback, interval_s=self._acquisition_interval_s)
        self._daq_simulator.start()

        return zmq_json_response(
            {
                "success": True,
                "message": {
                    "running": True,
                    "interval": self._acquisition_interval_s,
                    "already_running": False,
                },
            }
        )

    async def _do_stop_acquisition(self, cmd: dict[str, Any]) -> list:
        """Stop acquisition and return status/counter information."""
        was_running = await self.stop_acquisition()
        return zmq_json_response(
            {
                "success": True,
                "message": {
                    "running": False,
                    "stopped": was_running,
                    "acquisition logged": self._acquisition_logged_count,
                },
            }
        )

    async def stop_acquisition(self) -> bool:
        """Stop the simulator thread without blocking the event loop.

        Returns:
            True if acquisition was running and has been stopped, otherwise False.
        """
        simulator = self._daq_simulator
        if simulator is None:
            return False

        self._daq_simulator = None
        await asyncio.to_thread(simulator.stop)
        return True

    @override
    async def handle_acquisition_record(self, record: dict[str, Any]):
        """Receive one processed acquisition record and log it."""
        self._acquisition_logged_count += 1
        self.logger.info(f"Dummy acquisition record {self._acquisition_logged_count}: {record.get('data')}")

    async def _handle_health(self, cmd: dict[str, Any]) -> list:
        """Return a compact health payload for monitoring and tests."""
        return zmq_json_response(
            {
                "success": True,
                "message": {
                    "status": "ok",
                    "echo count": self._echo_count,
                    "last value": self._last_value,
                    "acquisition running": self._is_acquisition_running(),
                    "acquisition logged": self._acquisition_logged_count,
                },
            }
        )

    @override
    def stop(self):
        """Stop acquisition first, then stop the base server lifecycle."""
        if self._is_acquisition_running():
            if self._loop is not None and self._loop.is_running():
                self._loop.create_task(self.stop_acquisition())
            elif self._daq_simulator is not None:
                self._daq_simulator.stop()
                self._daq_simulator = None
        super().stop()


class DummyAsyncControlClient(TypedAsyncControlClient):
    """Example client wrapper with strongly named methods for dummy commands."""

    service_type = DummyAsyncControlServer.service_type

    async def echo(self, message: str, timeout: float | None = None) -> str | None:
        """Send an `echo` device command and return the echoed message."""
        response = await self.send_device_command({"command": "echo", "message": message}, timeout=timeout)
        return self._success_message_as_str(response, "echo")

    async def set_value(self, value: str, timeout: float | None = None) -> str | None:
        """Send `set-value` and return the stored value from the device response."""
        response = await self.send_device_command({"command": "set-value", "value": value}, timeout=timeout)
        message = self._success_message_as_dict(response, "set-value")
        if message is None:
            return None
        return str(message.get("stored"))

    async def health(self, timeout: float | None = None) -> dict[str, Any] | None:
        """Return the dummy server health payload."""
        response = await self.send_service_command("health", timeout=timeout)
        return self._success_message_as_dict(response, "health")

    async def start_acquisition(self, interval: float = 0.05, timeout: float | None = None) -> dict[str, Any] | None:
        """Start simulator-backed acquisition with the requested sample interval."""
        response = await self.send_device_command(
            {"command": "start-acquisition", "interval": interval},
            timeout=timeout,
        )
        return self._success_message_as_dict(response, "start-acquisition")

    async def stop_acquisition(self, timeout: float | None = None) -> dict[str, Any] | None:
        """Stop simulator-backed acquisition and return final acquisition status."""
        response = await self.send_device_command({"command": "stop-acquisition"}, timeout=timeout)
        return self._success_message_as_dict(response, "stop-acquisition")


app = typer.Typer()


@app.command(cls=TyperAsyncCommand)
async def start_cs():
    """Start the dummy async control server on localhost."""

    with remote_logging():
        try:
            control_server = DummyAsyncControlServer()
            await control_server.start()
        except KeyboardInterrupt:
            print("Shutdown requested...exiting")
        except SystemExit as exit_code:
            print(f"System Exit with code {exit_code}.")
            sys.exit(-1)
        except Exception:  # noqa
            import traceback

            traceback.print_exc(file=sys.stdout)


@app.command(cls=TyperAsyncCommand)
async def stop_cs():
    """Send a quit service command to the dummy async control server."""
    async with DummyAsyncControlClient() as dummy:
        logger.info("Sending a stop_server() to the async dummy control server.")
        await dummy.stop_server()


if __name__ == "__main__":
    app()



from __future__ import annotations

import asyncio
import datetime
import json
import pickle
import uuid
from asyncio import Event
from asyncio import Task
from enum import Enum
from enum import auto
from typing import Any
from typing import Callable
from typing import Coroutine

import zmq.asyncio
from rich.traceback import Traceback

from egse.env import bool_env
from egse.exceptions import InitializationError
from egse.log import logging

# from egse.process import ProcessStatus
from egse.registry.client import AsyncRegistryClient
from egse.system import Periodic
from egse.system import get_current_location
from egse.system import get_host_ip
from egse.system import humanize_seconds
from egse.system import log_rich_output
from egse.system import type_name
from egse.zmq_ser import get_port_number
from egse.zmq_ser import set_address_port
from egse.zmq_ser import zmq_error_response
from egse.zmq_ser import zmq_json_request
from egse.zmq_ser import zmq_json_response
from egse.zmq_ser import zmq_string_request
from egse.zmq_ser import zmq_string_response

logger = logging.getLogger("egse.async_control")

# When zero (0) ports will be dynamically allocated by the system
CONTROL_SERVER_DEVICE_COMMANDING_PORT = 0
CONTROL_SERVER_SERVICE_COMMANDING_PORT = 0

CONTROL_SERVER_SERVICE_TYPE = "async-control-server"
CONTROL_CLIENT_ID = "async-control-client"

VERBOSE_DEBUG: bool = bool_env("VERBOSE_DEBUG")

RECREATE_SOCKET = False
"""Recreate ZeroMQ socket after a timeout. Set to True when experiencing problems
with corrupted socket states. This is mainly used for REQ-REP protocols."""


class SocketType(Enum):
    """The socket type defines which socket to use for the intended communication."""

    DEVICE = auto()
    SERVICE = auto()


async def is_control_server_active(service_type: str, timeout: float = 0.5) -> bool:
    """
    Checks if the Control Server is running.

    This function sends a *Ping* message to the Control Server and expects a *Pong* answer back within the timeout
    period.

    Args:
        service_type (str): the service type of the control server to check
        timeout (float): Timeout when waiting for a reply [s, default=0.5]

    Returns:
        True if the Control Server is running and replied with the expected answer; False otherwise.
    """

    # I have a choice here to check if the control server is active/healthy.
    #
    # 1. I can connect to the ServiceRegistry, and analyse the 'health' field if it is 'passing', or

    with AsyncRegistryClient() as registry:
        service = await registry.discover_service(service_type)
    if service:
        return True if service["health"] == "passing" else False
    else:
        return False

    # 2. I can connect to the control server (by first contacting the service registry) and send a ping.

    # try:
    #     async with AsyncControlClient(service_type=service_type) as client:
    #         response = await client.ping()
    #     return response == "pong"
    # except InitializationError:
    #     return False


class AsyncControlServer:
    service_type = CONTROL_SERVER_SERVICE_TYPE

    def __init__(self):
        self.interrupted: Event = asyncio.Event()
        self.logger = logging.getLogger("egse.async_control.server")
        self._loop: asyncio.AbstractEventLoop | None = None
        self._start_time = datetime.datetime.now()

        self.mon_delay = 1000
        """Delay between publish status information [ms]."""
        self.hk_delay = 1000
        """Delay between saving housekeeping information [ms]."""

        # self._process_status = ProcessStatus()

        self._service_id = None

        self._sequential_queue: asyncio.Queue[Coroutine[Any, Any, Any]] = asyncio.Queue()
        """Queue for sequential operations that must preserve order of execution."""

        self.device_command_port = CONTROL_SERVER_DEVICE_COMMANDING_PORT
        """The device commanding port for the control server. This will be 0 at start and dynamically assigned by the
        system."""

        self.service_command_port = CONTROL_SERVER_SERVICE_COMMANDING_PORT
        """The service commanding port for the control server. This will be 0 at start and dynamically assigned by the
        system."""

        self.device_command_handlers: dict[str, Callable] = {}
        """Dictionary mapping device command names to their handler functions."""

        self.service_command_handlers: dict[str, Callable] = {}
        """Dictionary mapping service command names to their handler functions."""

        self._tasks: list[Task] = []
        """The background top-level tasks that are performed by the control server."""

        self._ctx = zmq.asyncio.Context.instance()

        # Socket to handle device commanding pattern - ROUTER-DEALER
        self.device_command_socket: zmq.asyncio.Socket = self._ctx.socket(zmq.ROUTER)

        # Socket to handle service commanding pattern - ROUTER-DEALER
        self.service_command_socket: zmq.asyncio.Socket = self._ctx.socket(zmq.ROUTER)

        self.register_default_device_command_handlers()
        self.register_default_service_command_handlers()

        # Call the hook for registering custom handlers, so that they are registered before the server
        # starts accepting commands.
        self.register_custom_handlers()

        self.registry = AsyncRegistryClient()

    def register_default_device_command_handlers(self):
        """Register baseline device handlers that are useful for diagnostics and examples."""
        self.add_device_command_handler("block", self._do_block)
        self.add_device_command_handler("say", self._do_say)

    def register_default_service_command_handlers(self):
        """Register baseline service handlers for lifecycle and health checks."""
        self.add_service_command_handler("terminate", self._handle_terminate)
        self.add_service_command_handler("info", self._handle_info)
        self.add_service_command_handler("ping", self._handle_ping)
        self.add_service_command_handler("block", self._handle_block)

    def register_custom_handlers(self):
        """Hook for subclasses to register device-specific command handlers."""

    def get_service_info(self) -> dict[str, Any]:
        """Service info payload returned by the `info` service command."""
        return {
            "name": type(self).__name__,
            "hostname": self.get_ip_address(),
            "device commanding port": self.device_command_port,
            "service commanding port": self.service_command_port,
            "service type": self.service_type,
            "uptime": humanize_seconds(
                (datetime.datetime.now() - self._start_time).total_seconds(), include_micro_seconds=False
            ),
        }

    @staticmethod
    def get_ip_address() -> str:
        """Returns the IP address of the current host or localhost."""
        return get_host_ip() or "localhost"

    def connect_device_command_socket(self):
        self.device_command_socket.bind(f"tcp://*:{self.device_command_port}")
        self.device_command_port = get_port_number(self.device_command_socket)

    def connect_service_command_socket(self):
        self.service_command_socket.bind(f"tcp://*:{self.service_command_port}")
        self.service_command_port = get_port_number(self.service_command_socket)

    def stop(self):
        self.logger.warning(f"Stopping the async control server {type(self).__name__}.")
        self.interrupted.set()

    async def start(self):
        self._loop = asyncio.get_running_loop()

        self.connect_device_command_socket()
        self.connect_service_command_socket()

        await self.register_service()

        self._tasks = self._create_background_tasks()

        try:
            while not self.interrupted.is_set():
                await self._check_tasks_health()
                await asyncio.sleep(1.0)
        except asyncio.CancelledError:
            self.logger.debug(f"Caught CancelledError on server keep-alive loop, terminating {type(self).__name__}.")
        finally:
            await self._cleanup_running_tasks()

        await self.deregister_service()

        self.disconnect_device_command_socket()
        self.disconnect_service_command_socket()

    def _create_background_tasks(self) -> list[Task]:
        """Create top-level server tasks.

        Subclasses can override this method and append extra tasks while keeping
        the base command/service lifecycle unchanged.
        """
        return [
            asyncio.create_task(self.process_device_command(), name="process-device-commands"),
            asyncio.create_task(self.process_service_command(), name="process-service-commands"),
            asyncio.create_task(self.send_status_updates(), name="send-status-updates"),
            asyncio.create_task(self.process_sequential_queue(), name="process-sequential-queue"),
        ]

    async def register_service(self):
        self.logger.info(f"Registering service {type(self).__name__} as type {self.service_type}")

        self.registry.connect()

        self._service_id = await self.registry.register(
            name=type_name(self),
            host=get_host_ip() or "127.0.0.1",
            port=get_port_number(self.device_command_socket),
            service_type=self.service_type,
            metadata={"service_port": get_port_number(self.service_command_socket)},
        )
        await self.registry.start_heartbeat()

    async def deregister_service(self):
        await self.registry.stop_heartbeat()
        await self.registry.deregister()

        self.registry.disconnect()

    async def _check_tasks_health(self):
        """Check if any tasks unexpectedly terminated."""
        for task in self._tasks:
            if task.done() and not task.cancelled():
                try:
                    # This will raise any exception that occurred in the task
                    task.result()
                except Exception as exc:
                    self.logger.error(f"Task {task.get_name()} failed: {exc}", exc_info=True)
                    # Potentially restart the task or shut down service

    async def _cleanup_running_tasks(self):
        # Cancel all running tasks
        for task in self._tasks:
            if not task.done():
                self.logger.debug(f"Cancelling task {task.get_name()}.")
                task.cancel()

        # Wait for tasks to complete their cancellation
        if self._tasks:
            try:
                await asyncio.gather(*self._tasks, return_exceptions=True)
            except asyncio.CancelledError as exc:
                self.logger.debug(f"Caught {type_name(exc)}: {exc}.")
                pass

    def disconnect_device_command_socket(self):
        self.logger.debug("Cleaning up device command sockets.")
        if self.device_command_socket:
            self.device_command_socket.close(linger=100)

    def disconnect_service_command_socket(self):
        self.logger.debug("Cleaning up service command sockets.")
        if self.service_command_socket:
            self.service_command_socket.close(linger=100)

    async def process_device_command(self):
        self.logger.info("Starting device command processing ...")

        while not self.interrupted.is_set():
            try:
                # Wait for a request with timeout to allow checking if still running
                try:
                    parts = await asyncio.wait_for(self.device_command_socket.recv_multipart(), timeout=1.0)
                except asyncio.TimeoutError:
                    continue

                if VERBOSE_DEBUG:
                    self.logger.debug(f"Received multipart message: {parts}")

                # For commanding, we only accept simple commands as a string or a complex command with arguments as
                # JSON data. In both cases, there are only three parts in this multipart message.
                client_id, sequence_id, message_type, data = parts
                if message_type == b"MESSAGE_TYPE:STRING":
                    device_command = {"command": data.decode("utf-8")}
                elif message_type == b"MESSAGE_TYPE:JSON":
                    device_command = json.loads(data.decode())
                else:
                    filename, lineno, function_name = get_current_location()
                    # We have an unknown message format, send an error message back
                    message = zmq_error_response(
                        {
                            "success": False,
                            "message": f"Incorrect message type: {message_type}",
                            "metadata": {
                                "data": data.decode(),
                                "file": filename,
                                "lineno": lineno,
                                "function": function_name,
                            },
                        }
                    )
                    await self.device_command_socket.send_multipart([client_id, sequence_id, *message])
                    continue

                self.logger.debug(f"Received device command: {device_command}")

                self.logger.debug("Process the command...")
                response = await self._process_device_command(device_command)

                self.logger.debug("Send the response...")
                await self.device_command_socket.send_multipart([client_id, sequence_id, *response])

            except asyncio.CancelledError:
                self.logger.debug("Device command handling task cancelled.")
                break

    async def _process_device_command(self, cmd: dict[str, Any]) -> list:
        command = cmd.get("command")
        if not command:
            return zmq_error_response(
                {
                    "success": False,
                    "message": "no command field provide, don't know what to do.",
                }
            )

        handler = self.device_command_handlers.get(command)
        if not handler:
            filename, lineno, function_name = get_current_location()
            return zmq_error_response(
                {
                    "success": False,
                    "message": f"Unknown command: {command}",
                    "metadata": {"file": filename, "lineno": lineno, "function": function_name},
                }
            )

        return await handler(cmd)

    def add_device_command_handler(self, command_name: str, command_handler: Callable):
        self.device_command_handlers[command_name] = command_handler

    def add_service_command_handler(self, command_name: str, command_handler: Callable):
        self.service_command_handlers[command_name] = command_handler

    async def _do_say(self, cmd: dict[str, Any]) -> list:
        self.logger.debug(f"Executing command: '{cmd['command']}")
        self.logger.debug(f"Message: {cmd['message']}")
        return zmq_string_response(f"Message said: {cmd['message']}")

    async def _do_block(self, cmd: dict[str, Any]) -> list:
        self.logger.debug(f"Blocking the commanding for {cmd['sleep']}s...")
        await asyncio.sleep(cmd["sleep"])
        self.logger.debug(f"Blocking finished after {cmd['sleep']}s.")
        return zmq_string_response("block: ACK")

    async def _handle_block(self, cmd: dict[str, Any]) -> list:
        self.logger.debug(f"Handling '{cmd['command']}' service request.")
        await asyncio.sleep(cmd["sleep"])
        self.logger.debug(f"Blocking finished after {cmd['sleep']}s.")
        return zmq_string_response("block: ACK")

    async def _handle_ping(self, cmd: dict[str, Any]) -> list:
        self.logger.debug(f"Handling '{cmd['command']}' service request.")
        return zmq_string_response("pong")

    async def _handle_info(self, cmd: dict[str, Any]) -> list:
        self.logger.debug(f"Handling '{cmd['command']}' service request.")
        return zmq_json_response(
            {
                "success": True,
                "message": self.get_service_info(),
            }
        )

    async def _handle_terminate(self, cmd: dict[str, Any]) -> list:
        self.logger.debug(f"Handling '{cmd['command']}' request.")

        self.stop()

        return zmq_json_response(
            {
                "success": True,
                "message": {"status": "terminating"},
            }
        )

    async def process_service_command(self):
        self.logger.info("Starting service command processing ...")

        while not self.interrupted.is_set():
            try:
                # Wait for a request with timeout to allow checking if still running
                try:
                    parts = await asyncio.wait_for(self.service_command_socket.recv_multipart(), timeout=1.0)
                except asyncio.TimeoutError:
                    continue

                if VERBOSE_DEBUG:
                    self.logger.debug(f"{parts=}")

                # For commanding, we only accept simple commands as a string or a complex command with arguments as
                # JSON data. In both cases, there are only three parts in this multipart message.
                client_id, sequence_id, message_type, data = parts
                if message_type == b"MESSAGE_TYPE:STRING":
                    service_command = {"command": data.decode("utf-8")}
                elif message_type == b"MESSAGE_TYPE:JSON":
                    service_command = json.loads(data.decode())
                else:
                    filename, lineno, function_name = get_current_location()
                    # We have an unknown message format, send an error message back
                    message = zmq_error_response(
                        {
                            "success": False,
                            "message": f"Incorrect message type: {message_type}",
                            "metadata": {
                                "data": data.decode(),
                                "file": filename,
                                "lineno": lineno,
                                "function": function_name,
                            },
                        }
                    )
                    await self.service_command_socket.send_multipart([client_id, sequence_id, *message])
                    continue

                self.logger.debug(f"Received service request: {service_command}")

                self.logger.debug("Process the command...")
                response = await self._process_service_command(service_command)

                self.logger.debug("Send the response...")
                await self.service_command_socket.send_multipart([client_id, sequence_id, *response])

            except asyncio.CancelledError:
                self.logger.debug("Service command handling task cancelled.")
                break

    async def _process_service_command(self, cmd: dict[str, Any]) -> list:
        command = cmd.get("command")
        if not command:
            return zmq_error_response(
                {
                    "success": False,
                    "message": "no command field provide, don't know what to do.",
                }
            )

        handler = self.service_command_handlers.get(command)
        if not handler:
            filename, lineno, function_name = get_current_location()
            return zmq_error_response(
                {
                    "success": False,
                    "message": f"Unknown command: {command}",
                    "metadata": {"file": filename, "lineno": lineno, "function": function_name},
                }
            )

        return await handler(cmd)

    async def process_sequential_queue(self):
        """
        Process operations that need to be executed sequentially.

        When the operation return "Quit" the processing is interrupted.
        """

        self.logger.info("Starting sequential queue processing ...")

        while not self.interrupted.is_set():
            try:
                operation = await asyncio.wait_for(self._sequential_queue.get(), 0.1)
                await operation
                self._sequential_queue.task_done()
            except asyncio.TimeoutError:
                continue
            except asyncio.CancelledError:
                break
            except Exception as exc:
                self.logger.error(f"Error processing sequential operation: {exc}")

    async def send_status_updates(self):
        """
        Send status information about the control server and the device connection to the monitoring channel.
        """

        self.logger.info("Starting status updates ...")

        async def status():
            self.logger.info(f"{datetime.datetime.now()} Sending status updates.")
            await asyncio.sleep(0.5)  # ideally, should not be larger than periodic interval

        try:
            periodic = Periodic(interval=1.0, callback=status)
            periodic.start()

            await self.interrupted.wait()

            periodic.stop()

        except asyncio.CancelledError:
            self.logger.debug("Caught CancelledError on status updates keep-alive loop.")

    async def enqueue_sequential_operation(self, operation: Coroutine[Any, Any, Any]):
        """
        Add an operation to the sequential queue.

        Args:
            operation: A coroutine object to be executed sequentially.
        """

        if self._sequential_queue is not None:  # sanity check
            self._sequential_queue.put_nowait(operation)

    async def _execute_sequential(self, operation: Coroutine[Any, Any, Any]) -> Any:
        """Execute one coroutine through the sequential queue and await its result.

        Use this helper from command handlers, service handlers, and internal tasks
        when all hardware-facing work must be serialized through one lane.
        """

        loop = asyncio.get_running_loop()
        result_future: asyncio.Future[Any] = loop.create_future()

        async def wrapped_operation():
            try:
                result = await operation
            except Exception as exc:
                if not result_future.done():
                    result_future.set_exception(exc)
                return

            if not result_future.done():
                result_future.set_result(result)

        try:
            await self.enqueue_sequential_operation(wrapped_operation())
        except Exception:
            operation.close()
            raise

        return await result_future


class AcquisitionAsyncControlServer(AsyncControlServer):
    """Async control server variant with callback-based acquisition queueing support."""

    def __init__(self):
        super().__init__()

        self.acquisition_queue_maxsize = 10_000
        """Maximum number of acquisition records buffered in memory."""

        self._acquisition_queue: asyncio.Queue[dict[str, Any]] = asyncio.Queue(maxsize=self.acquisition_queue_maxsize)
        """Queue for acquisition records received through callback ingestion."""

        self.acquisition_dropped_count = 0
        """Number of acquisition records dropped because the queue was full."""

        self.acquisition_batch_enabled = False
        """Enable grouped acquisition processing for higher-rate producers."""

        self.acquisition_batch_max_size = 100
        """Maximum number of records grouped into one batch before sink dispatch."""

        self.acquisition_batch_max_wait_s = 0.05
        """Maximum wait time [s] to collect records into a batch."""

    def get_service_info(self) -> dict[str, Any]:
        """Service info payload including acquisition queue statistics."""
        info = super().get_service_info()
        info.update(
            {
                "acquisition queue size": self._acquisition_queue.qsize(),
                "acquisition dropped": self.acquisition_dropped_count,
                "acquisition batch enabled": self.acquisition_batch_enabled,
                "acquisition batch size": self.acquisition_batch_max_size,
            }
        )
        return info

    def _create_background_tasks(self) -> list[Task]:
        tasks = super()._create_background_tasks()
        tasks.append(asyncio.create_task(self.process_acquisition_data(), name="process-acquisition-data"))
        return tasks

    def get_acquisition_callback(self) -> Callable[[Any], None]:
        """Return a callback that device drivers can call to ingest acquisition samples."""
        return self.on_acquisition_data

    def on_acquisition_data(
        self,
        data: Any,
        *,
        source: str | None = None,
        metadata: dict[str, Any] | None = None,
    ):
        """Ingest one acquisition record from any thread.

        This function is intentionally sync/thread-safe so it can be used as a
        device-driver callback.
        """

        record = {
            "timestamp": datetime.datetime.now(datetime.timezone.utc).isoformat(),
            "source": source,
            "data": data,
            "metadata": metadata or {},
        }

        loop = self._loop
        if loop is None or not loop.is_running():
            self.logger.warning("Dropping acquisition record because event loop is not running yet.")
            self.acquisition_dropped_count += 1
            return

        try:
            running_loop = asyncio.get_running_loop()
        except RuntimeError:
            running_loop = None

        if running_loop is loop:
            self._enqueue_acquisition_record(record)
        else:
            loop.call_soon_threadsafe(self._enqueue_acquisition_record, record)

    def _enqueue_acquisition_record(self, record: dict[str, Any]):
        try:
            self._acquisition_queue.put_nowait(record)
        except asyncio.QueueFull:
            self.acquisition_dropped_count += 1
            if self.acquisition_dropped_count % 1000 == 1:
                self.logger.warning(
                    "Acquisition queue full, dropping records "
                    f"(dropped={self.acquisition_dropped_count}, maxsize={self.acquisition_queue_maxsize})."
                )

    async def process_acquisition_data(self):
        """Drain acquisition records from the queue and dispatch them to sink hooks."""

        self.logger.info("Starting acquisition data processing ...")

        while not self.interrupted.is_set():
            try:
                if not self.acquisition_batch_enabled:
                    try:
                        record = await asyncio.wait_for(self._acquisition_queue.get(), timeout=0.1)
                    except asyncio.TimeoutError:
                        continue
                    batch = [record]
                else:
                    try:
                        first_record = await asyncio.wait_for(
                            self._acquisition_queue.get(), timeout=self.acquisition_batch_max_wait_s
                        )
                    except asyncio.TimeoutError:
                        continue

                    batch = [first_record]

                    # Drain additional records without blocking to create a compact, ordered batch.
                    while len(batch) < self.acquisition_batch_max_size:
                        try:
                            batch.append(self._acquisition_queue.get_nowait())
                        except asyncio.QueueEmpty:
                            break

                try:
                    await self.handle_acquisition_batch(batch)
                finally:
                    for _ in batch:
                        self._acquisition_queue.task_done()

            except asyncio.CancelledError:
                self.logger.debug("Acquisition data task cancelled.")
                break
            except Exception as exc:
                self.logger.error(f"Error processing acquisition record: {exc}", exc_info=True)

    async def handle_acquisition_batch(self, records: list[dict[str, Any]]):
        """Hook for batch sinks; default keeps strict sequential per-record processing."""
        for record in records:
            await self.handle_acquisition_record(record)

    async def handle_acquisition_record(self, record: dict[str, Any]):
        """Hook for subclasses to forward acquisition data to DB, storage manager, or other services."""
        self.logger.debug(
            f"Acquisition record received (source={record.get('source')}, queue={self._acquisition_queue.qsize()})."
        )


DEFAULT_CLIENT_REQUEST_TIMEOUT = 5.0  # seconds
"""Default timeout for sending requests to the control server."""
DEFAULT_LINGER = 100  # milliseconds
"""Default linger for ZeroMQ sockets."""


class AsyncControlClient:
    service_type: str | None = None

    def __init__(
        self,
        endpoint: str | None = None,
        service_type: str | None = None,
        client_id: str = CONTROL_CLIENT_ID,
        timeout: float = DEFAULT_CLIENT_REQUEST_TIMEOUT,
        linger: int = DEFAULT_LINGER,
    ):
        self.logger = logging.getLogger("egse.async_control.client")

        self.endpoint = endpoint
        self.service_type = service_type or self.__class__.service_type
        self.timeout = timeout  # seconds
        self.linger = linger  # milliseconds
        self._client_id = f"{client_id}-{uuid.uuid4()}".encode()
        self._sequence = 0

        self.context: zmq.asyncio.Context = zmq.asyncio.Context.instance()

        self.device_command_socket: zmq.asyncio.Socket | None = self.context.socket(zmq.DEALER)
        self.service_command_socket: zmq.asyncio.Socket | None = self.context.socket(zmq.DEALER)

        self.device_command_port: int = 0
        self.service_command_port: int = 0

        self._post_init_is_done = False

    async def _post_init(self) -> bool:
        """
        A post initialization method that sets the device and service commanding
        ports and the endpoint for this client. The information is retrieved from
        the service registry.

        Returns:
            The method returns True if the device and service commanding port and
            the endpoint could be determined from the service registry.

            If the post initialization step has already been called, a warning is
            issued, and the method returns True.

            If no service_type is known or the service_type is not registered,
            False is returned.

        """
        if self._post_init_is_done:
            if VERBOSE_DEBUG:
                self.logger.debug("The post_init function is already called, returning.")
            return True

        self._post_init_is_done = True
        if self.service_type:
            with AsyncRegistryClient() as registry:
                service = await registry.discover_service(self.service_type)
            if service:
                hostname = service["host"]
                self.device_command_port = port = service["port"]
                self.service_command_port = service["metadata"]["service_port"]
                self.endpoint = f"tcp://{hostname}:{port}"
                return True
            else:
                return False

        return False

    # Q: Why do we need this create method here?
    # A: The constructor (`__init__`) can not be an async method and to properly initialise the client,
    # we need to contact the ServiceRegistry for the hostname and port numbers. The service discovery
    # is an async operation.
    #
    # Additionally, it's not a good idea to perform such initialization inside the constructor of the
    # class anyway. The class can also be called as an async context manager, in which case the create()
    # is not needed and the post initialization will be done in the `__aenter__` method.

    @classmethod
    async def create(cls, service_type: str | None = None) -> AsyncControlClient:
        """Factory method that creates an AsyncControlClient and collects information about the service it needs to
        connect to."""
        resolved_service_type = service_type or cls.service_type
        if not resolved_service_type:
            raise InitializationError(
                "Could not initialise AsyncControlClient, no service_type provided and no class default configured."
            )

        client = cls(service_type=resolved_service_type)
        if not await client._post_init():
            raise InitializationError(
                "Could not initialise AsyncControlClient, "
                f"no service_type ({resolved_service_type}) found. "
                "Will not be able to connect to the control server."
            )
        return client

    async def send_device_command(
        self,
        request: dict[str, Any] | str,
        timeout: float | None = None,
    ) -> dict[str, Any]:
        """Send a command over the device command channel."""
        return await self._send_request(SocketType.DEVICE, request, timeout=timeout)

    async def send_service_command(
        self,
        request: dict[str, Any] | str,
        timeout: float | None = None,
    ) -> dict[str, Any]:
        """Send a command over the service command channel."""
        return await self._send_request(SocketType.SERVICE, request, timeout=timeout)

    def connect(self):
        self.connect_device_command_socket()
        self.connect_service_command_socket()

    def disconnect(self):
        self.disconnect_device_command_socket()
        self.disconnect_service_command_socket()

    def connect_device_command_socket(self):
        if self.endpoint is None:
            self.logger.warning("Cannot connect device command socket: endpoint is not defined.")
            return
        if self.device_command_socket is None:
            self.device_command_socket = self.context.socket(zmq.DEALER)
        self.device_command_socket.setsockopt(zmq.LINGER, self.linger)
        self.device_command_socket.setsockopt(zmq.IDENTITY, self._client_id)
        self.device_command_socket.connect(self.endpoint)

    def connect_service_command_socket(self):
        if self.endpoint is None:
            self.logger.warning("Cannot connect service command socket: endpoint is not defined.")
            return

        if self.service_command_port == 0:
            self.logger.warning("Service command port is 0 when connecting socket.")

        if self.service_command_socket is None:
            self.service_command_socket = self.context.socket(zmq.DEALER)
        self.service_command_socket.setsockopt(zmq.LINGER, self.linger)
        self.service_command_socket.setsockopt(zmq.IDENTITY, self._client_id)
        self.service_command_socket.connect(set_address_port(self.endpoint, self.service_command_port))

    def disconnect_device_command_socket(self):
        if self.device_command_socket:
            self.device_command_socket.setsockopt(zmq.LINGER, 100)
            self.device_command_socket.close()
            self.device_command_socket = None

    def disconnect_service_command_socket(self):
        if self.service_command_socket:
            self.service_command_socket.setsockopt(zmq.LINGER, 100)
            self.service_command_socket.close()
            self.service_command_socket = None

    async def __aenter__(self):
        if not self._post_init_is_done:
            if not await self._post_init():
                raise InitializationError(
                    f"Could not initialise AsyncControlClient, no service_type ({self.service_type}) found. Will not "
                    f"be able to connect to the control server."
                )
                return self

        self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    async def do(self, cmd: dict[str, Any], timeout: float = 5.0) -> dict[str, Any]:
        """
        Sends a device command to the control server and waits for a response.

        Args:
            cmd (dict): The command to send. Should contain at least a 'command' key.
            timeout (float, optional): Maximum time to wait for a response in seconds. Defaults to 5.0.

        Returns:
            dict: The response from the control server. Contains at least:
                - 'success' (bool): True if the command was processed successfully, False otherwise.
                - 'message' (str or dict): The server's response or an error message.

        Notes:
            If the request times out or a socket error occurs, the method attempts
            to reset the sockets and returns a response with 'success' set to False
            and an appropriate error message.
        """
        response = await self.send_device_command(cmd, timeout=timeout)
        return response

    async def block(self, sleep: int, timeout: int | None = None) -> str | None:
        cmd = {"command": "block", "sleep": sleep}
        response = await self.send_service_command(cmd, timeout=timeout)
        if response["success"]:
            return response["message"]
        else:
            self.logger.error(f"Server returned an error: {response['message']}")
            return None

    async def ping(self, timeout: int | None = None) -> str | None:
        response = await self.send_service_command("ping", timeout=timeout)
        if response["success"]:
            return response["message"]
        else:
            self.logger.error(f"Server returned an error: {response['message']}")
            return None

    async def info(self) -> str | None:
        response = await self.send_service_command("info")
        if response["success"]:
            return response["message"]
        else:
            self.logger.error(f"Server returned an error: {response['message']}")
            return None

    async def stop_server(self) -> dict | None:
        response = await self.send_service_command("terminate")
        if response["success"]:
            return response["message"]
        else:
            self.logger.error(f"Server returned an error: {response['message']}")
            return None

    async def _send_request(
        self,
        socket_type: SocketType,
        request: dict[str, Any] | str,
        timeout: float | None = None,
    ) -> dict[str, Any]:
        """
        Send a request to the control server and get the response.

        A request can be a string with a simple command, e.g. 'ping', or it can be a dictionary
        in which case it will be sent as a JSON request. The dictionary shall have the following format:

            request = {
                'command': <the command string without arguments>,
                'args': [*args],
                'kwargs': {**kwargs},
            }

        The response from the server will always be a dictionary with at least the following structure:

            response = {
                'success': <True or False>,
                'message': <The content of the data returned by the server>,
            }

        Args:
            socket_type: socket type to use for the command request.
            request: The request to send to the control server.
            timeout: how many seconds before the request times out.

        Returns:
            The response from the control server as a dictionary.
        """

        timeout = timeout or self.timeout

        socket = self._get_socket(socket_type)
        self._sequence += 1
        try:
            if socket is None:
                raise RuntimeError("Socket of the AsyncControlClient is not initialized.")

            if isinstance(request, str):
                message = zmq_string_request(request)
            elif isinstance(request, dict):
                message = zmq_json_request(request)
            else:
                raise ValueError(f"request argument shall be a string or a dictionary, not {type(request)}.")

            if VERBOSE_DEBUG:
                self.logger.debug(f"Sending multipart message: {message}")

            await socket.send_multipart([str(self._sequence).encode(), *message])

            while True:
                sequence, msg_type, data = await asyncio.wait_for(socket.recv_multipart(), timeout=timeout)

                if VERBOSE_DEBUG:
                    self.logger.debug(f"Received multipart message: {sequence}, {msg_type}, {data}")

                if int(sequence.decode()) < self._sequence:
                    self.logger.warning(
                        f"Received a reply from a previous command: {int(sequence.decode())}, {msg_type}, {data}, "
                        f"current sequence id = {self._sequence}"
                    )
                    continue
                elif int(sequence.decode()) > self._sequence:
                    self.logger.error(
                        f"Protocol violation: received reply for future request {sequence} "
                        f"(current sequence: {self._sequence}). "
                        f"Possible bug or multiple clients with the same identity."
                    )
                    raise RuntimeError(
                        f"Protocol violation: "
                        f"Got a reply with a future sequence id {sequence}, current sequence id is {self._sequence}."
                    )

                if msg_type == b"MESSAGE_TYPE:STRING":
                    return {"success": True, "message": data.decode("utf-8")}
                elif msg_type == b"MESSAGE_TYPE:JSON":
                    return json.loads(data)
                elif msg_type == b"MESSAGE_TYPE:ERROR":
                    return pickle.loads(data)
                else:
                    msg = f"Unknown server response message type: {msg_type}"
                    self.logger.error(msg)
                    return {"success": False, "message": msg}

        except asyncio.TimeoutError:
            self.logger.error(f"{socket_type.name} request timed out after {timeout:.3f}s")
            self.recreate_socket(socket_type)
            return {
                "success": False,
                "message": f"Request timed out after {timeout:.3f}s",
            }

        except zmq.ZMQError as exc:
            self.logger.error(f"ZMQ error: {exc}")
            self.recreate_socket(socket_type)
            return {
                "success": False,
                "message": f"ZMQError: {exc}",
            }

        except Exception as exc:
            self.logger.error(f"Error sending request: {type(exc).__name__} – {exc}")
            traceback = Traceback.from_exception(
                type(exc),
                exc,
                exc.__traceback__,
                show_locals=True,  # Optional: show local variables
                width=None,  # Optional: use full width
                extra_lines=3,  # Optional: context lines
            )
            log_rich_output(self.logger, logging.ERROR, traceback)
            return {"success": False, "message": str(exc)}

    def _get_socket(self, socket_type: SocketType):
        match socket_type:
            case SocketType.DEVICE:
                return self.device_command_socket
            case SocketType.SERVICE:
                return self.service_command_socket

    def recreate_socket(self, socket_type: SocketType):
        if RECREATE_SOCKET:
            self.logger.warning(f"Recreating {socket_type.name} socket ...")
            match socket_type:
                case SocketType.DEVICE:
                    self.disconnect_device_command_socket()
                    self.connect_device_command_socket()
                case SocketType.SERVICE:
                    self.disconnect_service_command_socket()
                    self.connect_service_command_socket()
        else:
            self.logger.debug("Socket recreation after timeout has been disabled.")


class TypedAsyncControlClient(AsyncControlClient):
    """Small helper base class for client subclasses with typed command wrappers.

    Subclasses can call these helpers to keep command methods concise and consistent.
    """

    def _success_message(self, response: dict[str, Any], command_name: str) -> Any | None:
        if response.get("success"):
            return response.get("message")
        self.logger.error(f"{command_name} failed: {response.get('message')}")
        return None

    def _success_message_as_str(self, response: dict[str, Any], command_name: str) -> str | None:
        message = self._success_message(response, command_name)
        if message is None:
            return None
        if isinstance(message, str):
            return message
        self.logger.error(f"{command_name} returned non-string message: {type(message).__name__}")
        return None

    def _success_message_as_dict(self, response: dict[str, Any], command_name: str) -> dict[str, Any] | None:
        message = self._success_message(response, command_name)
        if message is None:
            return None
        if isinstance(message, dict):
            return message
        self.logger.error(f"{command_name} returned non-dict message: {type(message).__name__}")
        return None


async def cs_test_device_command_timeouts():
    """
    This test is about timeouts on the client and blocking commands on the server
    for *device commands*.

    We start a control server which will register to the service registry and
    send out heartbeats.

    We start a control client that will connect to the server through its service type.
    Then we send the following commands, from the client perspective:

    - send a 'block' device command of 8s and a timeout of 3s. This will block the
      device commanding part on the server for ten seconds. The client will
      time out on this command after three seconds.
    - sleep for ten seconds.
    - send a 'say' command with a timeout of 2s. This will return immediately.

    The client should properly discard the server reply from the timed out block
    command. You should see a warning message saying a reply was received from a
    previous command. The 'say' command should receive its response as expected.

    """

    # First start the control server as a background task.
    server = AsyncControlServer()
    server_task = asyncio.create_task(server.start())

    logger.info("Starting Asynchronous Control Server ...")

    # Give the control server the time to start up
    logger.info("Sleep for 0.5s...")
    await asyncio.sleep(0.5)

    # As of now, the server_task is running 'in the background' in the event loop.

    # Now create a control client that will connect to the above server.
    async with AsyncControlClient(service_type=CONTROL_SERVER_SERVICE_TYPE) as client:
        # Sleep some time, so we can see the control server in action, e.g. status reports, housekeeping, etc
        logger.info("Sleep for 5s...")
        await asyncio.sleep(5.0)

        logger.info("Send a blocking device command, duration is 8s, timeout is 3s.")
        response = await client.do({"command": "block", "sleep": 8}, timeout=3)
        logger.info(f"block: {response=}")

        if response["success"] or "timed out" not in response["message"]:
            logger.error(f"Did not get expected message from server: {response['message']}")

        logger.info("Sleep for 10s...")
        await asyncio.sleep(10.0)

        logger.info("Send a 'say' device command, timeout is 2s.")
        response = await client.do({"command": "say", "message": "Hello, World!"}, timeout=2.0)
        logger.info(f"say: {response=}")

        if not response["success"] or "Hello, World!" not in response["message"]:
            logger.error(f"Did not get expected message from server: {response['message']}")

        is_active = await is_control_server_active(service_type=CONTROL_SERVER_SERVICE_TYPE)
        logger.info(f"Server status: {'active' if is_active else 'unreachable'}")

        logger.info("Sleeping 1s before terminating the server...")
        await asyncio.sleep(1.0)

        logger.info("Terminating the server.")
        response = await client.stop_server()
        logger.info(f"stop_server: {response = }")

    is_active = await is_control_server_active(service_type=CONTROL_SERVER_SERVICE_TYPE)
    logger.info(f"Server status: {'active' if is_active else 'unreachable'}")

    await server_task

    is_active = await is_control_server_active(service_type=CONTROL_SERVER_SERVICE_TYPE)
    logger.info(f"Server status: {'active' if is_active else 'unreachable'}")


async def cs_test_service_command_timeouts():
    """
    This test is about timeouts on the client and blocking commands on the server
    for *service commands*.

    We start a control server which will register to the service registry and
    send out heartbeats.

    We start a control client that will connect to the server through its service type.
    Then we send the following commands, from the client perspective:

    - send a 'block' service command of 8s and a timeout of 3s. This will block the
      service commanding part on the server for ten seconds. The client will
      time out on this command after three seconds.
    - sleep for ten seconds.
    - send a 'info' command with a timeout of 2s. This will return immediately.

    The client should properly discard the server reply from the timed out block
    command. You should see a warning message saying a reply was received from a
    previous command. The 'info' command should receive its response as expected.

    """

    # First start the control server as a background task.
    server = AsyncControlServer()
    server_task = asyncio.create_task(server.start())

    logger.info("Starting Asynchronous Control Server ...")

    # Give the control server the time to start up
    logger.info("Sleep for 0.5s...")
    await asyncio.sleep(0.5)

    # As of now, the server_task is running 'in the background' in the event loop.

    # Now create a control client that will connect to the above server.
    async with AsyncControlClient(service_type=CONTROL_SERVER_SERVICE_TYPE) as client:
        # Sleep some time, so we can see the control server in action, e.g. status reports, housekeeping, etc
        logger.info("Sleep for 5s...")
        await asyncio.sleep(5.0)

        logger.info("Send a blocking service command, duration is 8s, timeout is 3s.")
        response = await client.block(sleep=8, timeout=3)
        logger.info(f"service block: {response=}")

        logger.info("Sleep for 10s...")
        await asyncio.sleep(10.0)

        logger.info("Get info on the control server.")
        response = await client.info()
        logger.info(f"service info: {response=}")

        is_active = await is_control_server_active(service_type=CONTROL_SERVER_SERVICE_TYPE)
        logger.info(f"Server status: {'active' if is_active else 'unreachable'}")

        logger.info("Sleeping 1s before terminating the server...")
        await asyncio.sleep(1.0)

        logger.info("Terminating the server.")
        response = await client.stop_server()
        logger.info(f"stop_server: {response = }")

    is_active = await is_control_server_active(service_type=CONTROL_SERVER_SERVICE_TYPE)
    logger.info(f"Server status: {'active' if is_active else 'unreachable'}")

    await server_task

    is_active = await is_control_server_active(service_type=CONTROL_SERVER_SERVICE_TYPE)
    logger.info(f"Server status: {'active' if is_active else 'unreachable'}")


async def control_server_test():
    # First start the control server as a background task.
    server = AsyncControlServer()
    server_task = asyncio.create_task(server.start())

    logger.info("Starting Asynchronous Control Server ...")

    # Give the control server the time to start up
    logger.info("Sleep for 0.5s...")
    await asyncio.sleep(0.5)

    # As of now, the server_task is running 'in the background' in the event loop.

    # Now create a control client that will connect to the above server.
    async with AsyncControlClient(service_type=CONTROL_SERVER_SERVICE_TYPE) as client:
        # Sleep some time, so we can see the control server in action, e.g. status reports, housekeeping, etc
        logger.info("Sleep for 5s...")
        await asyncio.sleep(5.0)

        logger.info("Send a 'ping' service command...")
        response = await client.ping()
        logger.info(f"ping service command: {response = }")  # should be: response = 'pong'

        logger.info("Send an 'info' service command...")
        response = await client.info()
        logger.info(f"info service command: {response = }")  # should be: response = {'name': 'AsyncControlServer', ...

        logger.info("Send an 'info' device command...")
        # info() is a service command and not a device command, so this will fail.
        response = await client.do({"command": "info"})
        # should be: response={'success': False, 'message': 'Unknown command: info', ...
        logger.info(f"info device command: {response=}")

        # this will block commanding since device commands are executed in the same task
        # the do() will timeout after 3s, but the block command will run for 10s on the
        # server, and it will send back an ACK (which will be caught and interpreted by
        # –in this case– the seconds do 'say' command, which will not yet have timed out.
        logger.info("Send a blocking device command, duration is 9s, timeout is 3s.")
        response = await client.do({"command": "block", "sleep": 9}, timeout=3.0)
        # should be: response={'success': False, 'message': 'Request timed out after 3.000s'}
        logger.info(f"Blocking device command: {response=}")

        # ping() is a service command, so this is not blocked by the above block() command
        logger.info("Send a 'ping' service command...")
        response = await client.ping()
        logger.info(f"ping service command: {response=}")

        # say() is a device command and will time out after 2s because the above blocking
        # block() command is still running on the server
        logger.info("Send a 'say' device command, timeout is 2s.")
        response = await client.do({"command": "say", "message": "Hello, World!"}, timeout=2.0)
        logger.info(f"say device command: {response=}")

        # Sleep some time, so we can see the control server in action, e.g. status reports, housekeeping, etc
        logger.info("Sleep for 10s...")
        await asyncio.sleep(10.0)

        # This time we won't let the command time out...
        logger.info("Send a 'say' device command, timeout is 2s.")
        response = await client.do({"command": "say", "message": "Hello, Again!"}, timeout=2.0)
        logger.info(f"say device command: {response=}")

        is_active = await is_control_server_active(service_type=CONTROL_SERVER_SERVICE_TYPE)
        logger.info(f"Server status: {'active' if is_active else 'unreachable'}")

        logger.info("Sleeping 1s before terminating the server...")
        await asyncio.sleep(1.0)

        logger.info("Terminating the server.")
        response = await client.stop_server()
        logger.info(f"stop_server: {response = }")

    is_active = await is_control_server_active(service_type=CONTROL_SERVER_SERVICE_TYPE)
    logger.info(f"Server status: {'active' if is_active else 'unreachable'}")

    await server_task

    is_active = await is_control_server_active(service_type=CONTROL_SERVER_SERVICE_TYPE)
    logger.info(f"Server status: {'active' if is_active else 'unreachable'}")


if __name__ == "__main__":
    # The statement logging.captureWarnings(True) redirects Python's warnings module output
    # (such as warnings.warn(...)) to the logging system, so warnings appear in your logs
    # instead of just printing to stderr. This is useful for unified logging and easier
    # debugging, especially in larger applications.
    logging.captureWarnings(True)

    try:
        main_task = asyncio.run(cs_test_device_command_timeouts())
        # main_task = asyncio.run(cs_test_service_command_timeouts())
        # main_task = asyncio.run(control_server_test())
    except KeyboardInterrupt:
        print("Caught KeyboardInterrupt, terminating.")



"""
CommandProtocol is a base class for communicating commands with the hardware or
the control server. This class implements methods to send command messages and
receive responses.

The protocol also knows how to load the commands from the YAML file that contains
command definitions.

"""

from __future__ import annotations

__all__ = [
    "get_method",
    "get_function",
    "CommandProtocol",
    "DynamicCommandProtocol",
]

import abc
import inspect
import pickle
import types
from typing import Any
from typing import Type

from prometheus_client import Counter
from prometheus_client import Summary

from egse.command import Command
from egse.command import CommandExecution
from egse.control import ControlServer
from egse.decorators import deprecate
from egse.device import DeviceConnectionObserver
from egse.env import bool_env
from egse.log import logger
from egse.response import Failure
from egse.system import format_datetime
from egse.system import type_name

# Define some metrics for Prometheus

COMMAND_REQUESTS = Counter("cs_command_requests_count", "Count the number of commands", ["target"])
EXECUTION_TIME = Summary("cs_command_execution_time_seconds", "Time spent executing a command")

VERBOSE_DEBUG = bool_env("VERBOSE_DEBUG")


def get_method(parent_obj: object, method_name: str) -> types.MethodType | types.MethodWrapperType | None:
    """
    Returns a bound method from a given class *instance*.

    Args:
        parent_obj: the class instance that provides the method
        method_name: name of the method that is requested

    Returns:
        the method [type: method].

    Note:
        The method returned is a bound class instance method and therefore
        this method *does not* expects as its first argument the class
        instance, i.e. self, when you call this as a function.

    """
    if method_name in (None, "None", ""):
        return None

    if hasattr(parent_obj, method_name):
        method = getattr(parent_obj, method_name)
        if (
            inspect.ismethod(method)
            or isinstance(method, types.MethodWrapperType)
            or hasattr(method, "__method_wrapper")
        ):
            return method
        logger.warning(f"{method_name} is not a method, type={type(method)}")
    else:
        logger.warning(f"{parent_obj!r} has no method called {method_name}")

    return None


def get_function(parent_class: type, method_name: str) -> types.FunctionType | None:
    """
    Returns a function (unbound method) from a given class.

    Args:
        parent_class: the class that provides the method
        method_name: name of the method that is requested

    Returns:
        the method [type: function].

    Note:
        The function returned is an unbound class instance method and
        therefore this function expects as its first argument the class
        instance, i.e. self, when you call it as a function.

    """
    if method_name is None or method_name == "None":
        return None

    if hasattr(parent_class, method_name):
        func = getattr(parent_class, method_name)
        if inspect.isfunction(func):
            return func
        logger.warning(f"{method_name} is not a function, type={type(func)}")
    else:
        logger.warning(f"{parent_class.__module__}.{parent_class.__name__} has no method called {method_name}")

    return None


class BaseCommandProtocol(DeviceConnectionObserver):
    def __init__(self, control_server: ControlServer):
        super().__init__()
        self.__socket = None
        self._control_server = control_server

    def bind(self, socket):
        """Bind to a socket to listen for commands."""
        self.__socket = socket

        endpoint = self.get_bind_address()
        logger.info(f"Binding {type_name(self)} to {endpoint}")

        self.__socket.bind(endpoint)

    def get_bind_address(self):
        """
        Returns a string with the bind address, the endpoint, for accepting connections
        and bind a socket to.

        The port to connect to depends on the protocol implementation and therefore this
        method shall be implemented by the subclass.

        Returns:
            a string with the protocol and port to bind a socket to.
        """
        raise NotImplementedError(f"The get_bind_address() method shall be implemented for {type(self).__name__}.")

    def is_alive(self) -> bool:
        """
        This method can be overridden by a subclass to check whether any Thread or sub-process
        that was started is still alive.
        """
        # Don't make this a static method, because the subclass can still need access to `self`.
        return True

    @deprecate(alternative="the property")
    def get_control_server(self):
        """
        Return the control server to which this protocol is associated.
        """
        return self.control_server

    @property
    def control_server(self):
        """
        Return the control server to which this protocol is associated.
        """
        return self._control_server

    def get_status(self):
        """
        Returns a dictionary with status information for the control server, enhanced by the
        subclass with device specific status information.

        This method should be implemented/overridden by the sub-class. The sub-class specific
        method should update the dictionary returned by this super-class method with device
        specific status values.

        The dict returned by this method includes the following keywords:

        * timestamp (str): a string representation of the current datetime
        * PID (int): the Process ID for the control server
        * Up (float): the uptime of the control server [s]
        * UUID (uuid1): a UUID for the control server
        * RSS (int): the 'Resident Set Size', this is the non-swapped physical memory a process
            has used [byte]
        * USS (int): the 'Unique Set Size', this is the memory which is unique to a process [byte]
        * CPU User (float): time spent in user mode [s]
        * CPU System (float): time spent in kernel mode [s]
        * CPU% (float): the process CPU utilization as a percentage [%]

        Check the documentation for `psutil.Process` for more in-depth information about the
        dict keys.

        Returns:
            a dictionary with status information.
        """
        status = {
            "timestamp": format_datetime(),
            "delay": self._control_server.mon_delay,
        }
        status.update(self._control_server.get_process_status())
        return status

    def get_housekeeping(self) -> dict:
        """Returns a dictionary with housekeeping information about the device."""
        raise NotImplementedError(f"The get_housekeeping() method shall be implemented for {self.__class__.__name__}.")

    def get_device(self):
        """Returns the device object for the device that is controlled by this protocol."""
        raise NotImplementedError(f"The get_device() method shall be implemented for {self.__class__.__name__}.")

    def send(self, data):
        """
        Send a message to the ControlServer. The message shall be fully populated
        and is only serialized before sending over the ZeroMQ socket.

        FIXME: We need to add error handling here, e.g. what if the send() fails? Do we need
               to implement retries as with Proxy?
        """
        pickle_string = pickle.dumps(data)
        self.__socket.send(pickle_string)

    def receive(self):
        """
        Receive a serialized message from the ControlServer. The message will not
        be decoded/de-serialized, but is returned as it was sent. Decoding shall
        be handled by the calling method.
        """
        pickle_string = self.__socket.recv()
        return pickle.loads(pickle_string)

    def send_commands(self):
        """
        This method will be implemented by the subclass when the protocol needs to
        send the command definitions to the client. This is the case for the
        CommandProtocol where the commands are decorated with `@dynamic_interface`
        but not for the DynamicCommandProtocol where the commands are decorated
        with `@dynamic_command`.
        """
        ...

    # FIXME:
    #   We might want to reconsider how commands are send over the ZeroMQ sockets.
    #   it can be very useful to use multipart messages here with the type and
    #   origin etc. to ease the if..else.. constructs.

    @EXECUTION_TIME.time()
    def execute(self):
        cs = self.get_control_server()
        data = self.receive()
        cmd = None
        args = kwargs = None
        if isinstance(data, CommandExecution):
            cmd = data.get_cmd()
            cmd_name = cmd.get_name()
            args = data.get_args()
            kwargs = data.get_kwargs()
        elif isinstance(data, dict):
            cmd_name = data.get("cmd")
            args = data.get("args")
            kwargs = data.get("kwargs")
        elif isinstance(data, str):
            cmd_name = data
        else:
            cmd_name = None

        logger.log(0, f"cmd_name = {cmd_name}")

        # Server availability request - Ping-Pong

        if cmd_name == "Ping":
            COMMAND_REQUESTS.labels(target="ping").inc()
            self.send("Pong")
        elif cmd_name == "send_commands":
            self.send_commands()
        elif cmd_name == "get_service_port":
            self.send(cs.get_service_port())
        elif cmd_name == "get_monitoring_port":
            self.send(cs.get_monitoring_port())
        elif cmd_name == "get_commanding_port":
            self.send(cs.get_commanding_port())
        elif cmd_name == "get_ip_address":
            self.send(cs.get_ip_address())
        elif cmd_name == "get_storage_mnemonic":
            self.send(cs.get_storage_mnemonic())
        elif cmd:
            COMMAND_REQUESTS.labels(target="device").inc()
            cmd.server_call(self, *args, **kwargs)
        else:
            COMMAND_REQUESTS.labels(target="invalid").inc()
            logger.warning(f"Invalid command received: {cmd_name}")
            self.send(Failure(f"Invalid command: {cmd_name}"))

    def handle_device_method(self, cmd: Command, *args: list, **kwargs: dict):
        """
        Call the device method with the given arguments. This method is called at the server side.

        Args:
            cmd: the devices command class that knows which device command shall be called
            args: the arguments that will be passed on to the device command
            kwargs: the keyword arguments that will be passed on to the device command
        """

        raise NotImplementedError(
            f"The method `handle_device_method(..) shall be implemented by the class {self.__class__.__name__}.`"
        )

    def quit(self):
        """This method can be overridden by a subclass to clean up and stop threads that it started."""

        logger.debug("The quit() method was called on the command protocol base class.")


class DynamicCommandProtocol(BaseCommandProtocol, metaclass=abc.ABCMeta):
    def __init__(self, control_server: ControlServer):
        super().__init__(control_server)

    def handle_device_method(self, cmd: Command, *args, **kwargs):
        """
        Call the device method with the given arguments.

        Args:
            cmd: the devices command class that knows which device command shall be called
            args: the arguments that will be passed on to the device command
            kwargs: the keyword arguments that will be passed on to the device command
        """
        # The lookup table contains object (bound) methods, so we do not have to
        # provide the 'self' argument anymore.

        method_name = cmd.get_device_method_name()
        method = get_method(self.get_device(), method_name)

        # We treat the get_response function special as it needs to send the ``cmd`` string
        # to the device we need to pass the processed cmd string into the method.

        try:
            if method_name == "get_response":
                device_cmd_string = cmd.get_cmd_string(*args, *kwargs)
                logger.log(5, f"Executing method {method.__name__}({device_cmd_string})")
                response = method(device_cmd_string)
            else:
                logger.log(5, f"Executing method {method.__name__}({args}, {kwargs})")
                response = method(*args, **kwargs)
        except Exception as exc:
            logger.exception(f"Executing {method_name} failed.")
            # Pass the exception on to the client as a Failure message
            response = Failure(f"Executing {method_name} failed: ", exc)

        # Enable the following message only when debugging, because this log message can become
        # very long for data storage commands.
        # logger.debug(f"handle_device_method: {device_name}({args}, {kwargs}) -> {response!s}")

        self.send(response)


class CommandProtocol(BaseCommandProtocol, metaclass=abc.ABCMeta):
    """
    This class is the glue between the control servers and the hardware
    controllers on one side, and between the control server and the connected
    proxy classes on the other side.

    The connection with the hardware controllers is when the ``execute()`` method
    calls the ``server_call()`` method of the command class.

    The connection with the proxy classes is when the ``client_call()`` method is added to the
    interface of the Proxy subclass (by the ``_add_commands()`` method).

    FIXME: Protocol is not used at the client side, i.e. the Proxy class.
    """

    def __init__(self, control_server: ControlServer):
        super().__init__(control_server)
        self._commands = {}  # variable is used by subclasses
        self._method_lookup = {}  # lookup table for device methods

    def send_commands(self):
        """
        Send the command definitions that were loaded for the specific device.
        """
        self.send(self._commands)

    def load_commands(self, command_settings: dict, command_class: Type[Command], device_class: Any):
        """
        Loads the command definitions from the given ``command_settings`` and builds an internal
        dictionary containing the command names as keys and the corresponding ``Command`` class
        objects as values.

        The ``command_settings`` is usually loaded from a YAML configuration file containing the
        command definitions for the device.

        Args:
            command_settings: a dictionary containing the command definitions for this device
            command_class: the type of command to create, a subclass of Command
            device_class: the type of the base device class from which the methods are loaded
        """
        for name in command_settings:
            command_settings_name = command_settings[name]
            if "cmd" in command_settings_name:
                cmd = command_settings_name["cmd"]
            else:
                cmd = ""

            if "description" in command_settings_name:
                description = command_settings_name["description"]
            else:
                description = None

            # The response field is the name of a function from the CommandProtocol class or a
            # sub-class. This function shall send a response back to the client (Proxy). That's
            # why this field is called response.
            # By convention we like that this method name would start with `handle_` so the we
            # can make a distinction between response commands and normal methods in Protocol.
            # Remember that response methods should send a reply back to the client (which will
            # be waiting for it..).
            # If no response field is given, then the `handle_device_method` will be called.

            if "response" in command_settings_name:
                response_method = get_function(self.__class__, command_settings_name["response"])
            else:
                response_method = get_function(self.__class__, "handle_device_method")

            # The device_method field is used in the `handle_device_method` to call the method on
            # the device class. That is the class that implements the DeviceInterface and is
            # usually called a Controller or a Simulator.
            #
            # If no device_name field is given, the name from the command_settings is used.

            if "device_method" in command_settings_name:
                device_method_name = command_settings_name["device_method"]
            else:
                device_method_name = name

            # check if the device_method exists in the device base class

            if device_method_name == "None":
                device_method = None
            else:
                device_method = get_function(device_class, device_method_name)

            logger.log(
                0,
                f"Creating {command_class.__module__}.{command_class.__name__}(name='{name}', "
                f"cmd='{cmd}', "
                f"response={response_method}, device_method={device_method})",
            )
            logger.debug(f"Creating {command_class.__name__} command with {name=}, {cmd=}, {device_method=}")

            self._commands[name] = command_class(
                name=name,
                cmd=cmd,
                response=response_method,
                description=description,
                device_method=device_method,
            )

    def build_device_method_lookup_table(self, device_obj: Any):
        """
        Fill the lookup table with device command methods that are bound to the device object.

        Args:
            device_obj: instance of a device command class
        """
        for cmd in self._commands.values():
            method_name = cmd.get_device_method_name()
            method = get_method(device_obj, method_name)
            if method is not None:
                self._method_lookup[method_name] = method

    def handle_device_method(self, cmd: Command, *args: list, **kwargs: dict):
        """
        Call the device method with the given arguments.

        Args:
            cmd: the devices command class that knows which device command shall be called
            args: the arguments that will be passed on to the device command
            kwargs: the keyword arguments that will be passed on to the device command
        """
        # The lookup table contains object (bound) methods, so we do not have to
        # provide the 'self' argument anymore.

        device_name = cmd.get_device_method_name()
        method = self._method_lookup[device_name]

        # We treat the get_response function special as it needs to send the ``cmd`` string
        # to the device we need to pass the processed cmd string into the method.

        rc = 0

        try:
            if device_name == "get_response":
                device_cmd_string = cmd.get_cmd_string(*args, *kwargs)
                logger.debug(f"Executing method {method.__name__}({device_cmd_string})")
                response = method(device_cmd_string)
            else:
                logger.debug(f"Executing method {method.__name__}({args}, {kwargs})")
                response = method(*args, **kwargs)
        except Exception as exc:
            logger.exception(f"Executing {device_name} failed.")
            # Pass the exception on to the client as a Failure message
            response = Failure(f"Executing {device_name} failed: ", exc)

            rc = -1  # indicating we got an exception when executing the command on the server

        # Enable the following message only when debugging, because this log message can become
        # very long for data storage commands.
        if VERBOSE_DEBUG:
            logger.debug(f"handle_device_method: {device_name}({args}, {kwargs}) -> {response!s}")

        self.send(response)

        return rc
